import logging
import os
import inspect


class Logger:
    def __init__(self, name, level=logging.INFO, log_file=None):
        if not log_file:
            log_file = os.path.join(os.getcwd(), "coauthor.log")
        self.logger = logging.getLogger(name)
        self.logger.setLevel(level)

        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

        self.log_file = log_file
        file_handler = logging.FileHandler(log_file)
        file_handler.setLevel(level)
        file_handler.setFormatter(formatter)
        self.logger.addHandler(file_handler)

        console_handler = logging.StreamHandler()
        console_handler.setLevel(level)
        console_handler.setFormatter(formatter)
        self.logger.addHandler(console_handler)

    def clean_log_file(self):
        if self.log_file and os.path.exists(self.log_file):
            os.remove(self.log_file)
            print(f"Log file '{self.log_file}' has been removed.")

    def debug(self, msg):
        frame = inspect.currentframe().f_back
        caller_module = frame.f_globals["__name__"]
        caller_function = frame.f_code.co_name
        self.logger.debug(f"[{caller_module}.{caller_function}] {msg}")

    def info(self, msg):
        frame = inspect.currentframe().f_back
        caller_module = frame.f_globals["__name__"]
        caller_function = frame.f_code.co_name
        self.logger.info(f"[{caller_module}.{caller_function}] {msg}")

    def warning(self, msg):
        frame = inspect.currentframe().f_back
        caller_module = frame.f_globals["__name__"]
        caller_function = frame.f_code.co_name
        self.logger.warning(f"[{caller_module}.{caller_function}] {msg}")

    def error(self, msg):
        frame = inspect.currentframe().f_back
        caller_module = frame.f_globals["__name__"]
        caller_function = frame.f_code.co_name
        self.logger.error(f"[{caller_module}.{caller_function}] {msg}")



# import os
# import tempfile
# from langchain.text_splitter import RecursiveCharacterTextSplitter
# from langchain_community.document_loaders.sitemap import SitemapLoader
# from langchain_community.vectorstores import SKLearnVectorStore
# from langchain_openai import OpenAIEmbeddings

# RAG_PROMPT = """You are an assistant for question-answering tasks.
# Use the following pieces of retrieved context to answer the latest question in the conversation.
# If you don't know the answer, just say that you don't know.
# The pre-existing conversation may provide important context to the question.
# Use three sentences maximum and keep the answer concise.
# Important: if you know / have the URL of the source document, add it to your answer

# Conversation: {conversation}
# Context: {context}
# Question: {question}
# Answer:"""


# def get_vector_db_retriever():
#     persist_path = os.path.join(tempfile.gettempdir(), "union.parquet")
#     embd = OpenAIEmbeddings()

#     # If vector store exists, then load it
#     if os.path.exists(persist_path):
#         vectorstore = SKLearnVectorStore(embedding=embd, persist_path=persist_path, serializer="parquet")
#         return vectorstore.as_retriever(lambda_mult=0)

#     # Otherwise, index LangSmith documents and create new vector store
#     ls_docs_sitemap_loader = SitemapLoader(web_path="https://docs.smith.langchain.com/sitemap.xml")
#     ls_docs = ls_docs_sitemap_loader.load()

#     text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(chunk_size=500, chunk_overlap=0)
#     doc_splits = text_splitter.split_documents(ls_docs)

#     vectorstore = SKLearnVectorStore.from_documents(
#         documents=doc_splits, embedding=embd, persist_path=persist_path, serializer="parquet"
#     )
#     vectorstore.persist()
#     return vectorstore.as_retriever(lambda_mult=0)



from openai import OpenAI
from pydantic import BaseModel, Field

client = OpenAI()

"""
This utility module provides functions and classes to evaluate the semantic similarity between text responses.
It utilizes OpenAI's language model to score the similarity between an expected response and an actual response
on a numerical scale from 1 to 10.
"""


class Similarity_Score(BaseModel):
    """
    A data model to represent the semantic similarity score.

    Attributes:
        similarity_score (int): A score between 1 and 10 representing the semantic similarity
                                between two responses, where 1 indicates the responses are unrelated
                                and 10 indicates they are identical in meaning.
    """

    similarity_score: int = Field(
        description="Semantic similarity score between 1 and 10, where 1 means unrelated and 10 means identical."
    )


def compare_semantic_similarity(inputs: dict, reference_outputs: dict, outputs: dict):
    """
    Compare the semantic similarity of a given response to a reference response using OpenAI's language model.

    Args:
        inputs (dict): A dictionary with the question or context prompting the responses.
        reference_outputs (dict): A dictionary with the expected or correct response.
        outputs (dict): A dictionary with the actual response generated for evaluation.

    Returns:
        dict: A dictionary containing the similarity score and the key 'similarity'.
              The score is an integer between 1 and 10, where 1 indicates the response
              is entirely unrelated to the reference, and 10 indicates a response identical in meaning.
    """
    input_question = inputs["question"]
    reference_response = reference_outputs["output"]
    run_response = outputs["output"]

    completion = client.beta.chat.completions.parse(
        model="gpt-4o",
        messages=[
            {
                "role": "system",
                "content": (
                    "You are a semantic similarity evaluator. Compare the meanings of two responses to a question, "
                    "Reference Response and New Response, where the reference is the correct answer, and we are trying to judge if the new response is similar. "
                    "Provide a score between 1 and 10, where 1 means completely unrelated, and 10 means identical in meaning."
                ),
            },
            {
                "role": "user",
                "content": f"Question: {input_question}\n Reference Response: {reference_response}\n Run Response: {run_response}",
            },
        ],
        response_format=Similarity_Score,
    )

    similarity_score = completion.choices[0].message.parsed
    return {"score": similarity_score.similarity_score, "key": "similarity"}



"""Module for exporting coauthor profiles.

This module provides functions to export predefined profiles, including configuration files and templates,
to the current working directory. It handles file copying, checks for existing files, and displays differences
if conflicts are found.
"""

import importlib.resources
import pathlib
import shutil
from difflib import unified_diff


def export(config, logger):
    """Export a coauthor profile based on the provided configuration.

    This function checks if a profile is specified in the configuration arguments. If not, it lists available
    profiles and returns early. Otherwise, it calls the export_profile function to perform the export.

    Args:
        config (dict): Configuration dictionary containing command-line arguments.
        logger (logging.Logger): Logger instance for logging messages.

    Returns:
        None: If no profile is provided or on successful export.
    """
    args = config["args"]

    if not args.profile:
        profiles_path = importlib.resources.files("coauthor.profiles")
        profiles = [p.name for p in profiles_path.iterdir() if p.is_dir()]
        logger.info("Please provide a profile using --profile, for example --profile <profile-name>")
        logger.info("Available profiles: " + ", ".join(profiles))
        return

    profile = args.profile
    export_profile(profile, logger)


def is_files_equal(source, target):
    """Check if two files have identical content.

    This function compares the content of two files by reading them in chunks of 8192 bytes.
    It returns True if the files are identical, False otherwise.

    Args:
        source (pathlib.Path): Path to the source file.
        target (pathlib.Path): Path to the target file.

    Returns:
        bool: True if the files are equal, False otherwise.
    """
    with source.open("rb") as fsrc, target.open("rb") as ftgt:
        while True:
            buffer1 = fsrc.read(8192)
            buffer2 = ftgt.read(8192)
            if buffer1 != buffer2:
                return False
            if not buffer1:
                return True


def export_file(source, target, logger):
    """Export a file from source to target, handling existing files.

    If the target file does not exist, it creates the necessary directories and copies the file.
    If the target exists, it checks if the content is identical. If identical, logs a message.
    If different, logs the differences using unified diff.

    Args:
        source (pathlib.Path): Path to the source file.
        target (pathlib.Path): Path to the target file.
        logger (logging.Logger): Logger instance for logging messages.
    """
    if not target.exists():
        target.parent.mkdir(parents=True, exist_ok=True)
        with source.open("rb") as fsrc, target.open("wb") as ftgt:
            shutil.copyfileobj(fsrc, ftgt)
        logger.info(f"Created {target}")
    else:
        if is_files_equal(source, target):
            logger.info(f"Already exported: {target}")
        else:
            logger.info(f"File exists but is different: {target}")
            with source.open() as fsrc, target.open() as ftgt:
                source_lines = fsrc.readlines()
                target_lines = ftgt.readlines()
            diff = unified_diff(target_lines, source_lines, fromfile=str(target), tofile=str(source))
            logger.info("Diff:\n" + "\n".join(diff))


def export_config(profile_path, logger):
    """Export the configuration file (config.yml) from the profile to the current directory.

    Copies config.yml to .coauthor.yml in the current working directory if it exists in the profile.
    Uses export_file to handle the copying and logging.

    Args:
        profile_path (pathlib.Path): Path to the profile directory.
        logger (logging.Logger): Logger instance for logging messages.
    """
    config_source = profile_path / "config.yml"
    if not config_source.is_file():
        logger.warning(f"config.yml not found in profile {profile_path.name}")
        return

    config_target = pathlib.Path.cwd() / ".coauthor.yml"
    export_file(config_source, config_target, logger)


def export_templates(profile_path, logger):
    """Export the templates directory from the profile to the current directory.

    Recursively walks through the templates directory, ignoring specified folders like __pycache__,
    and exports each file to .coauthor/templates in the current working directory.
    Uses export_file for individual file handling.

    Args:
        profile_path (pathlib.Path): Path to the profile directory.
        logger (logging.Logger): Logger instance for logging messages.
    """
    templates_source = profile_path / "templates"
    if not templates_source.is_dir():
        logger.warning(f"templates directory not found in profile {profile_path.name}")
        return

    ignored_folders = ["__pycache__"]

    def walk(trav):
        yield trav
        if trav.is_dir():
            for sub in trav.iterdir():
                if sub.is_dir() and sub.name in ignored_folders:
                    continue
                yield from walk(sub)

    target_base = pathlib.Path.cwd() / ".coauthor" / "templates"
    for source_path in walk(templates_source):
        if source_path.is_file():
            rel_path = source_path.relative_to(templates_source)
            target_path = target_base / rel_path
            export_file(source_path, target_path, logger)


def export_profile(profile, logger):
    """Export a specific coauthor profile.

    This function exports the configuration file (config.yml) and templates directory from the specified
    profile to the current working directory. It handles creation of new files, skips ignored folders,
    checks for file equality, and logs differences using unified diff if files differ.

    Args:
        profile (str): Name of the profile to export.
        logger (logging.Logger): Logger instance for logging messages.
    """
    logger.info(f"Exporting profile {profile}")

    profile_path = importlib.resources.files("coauthor.profiles") / profile
    if not profile_path.is_dir():
        logger.error(f"Profile {profile} not found")
        return

    export_config(profile_path, logger)
    export_templates(profile_path, logger)



import tempfile
import logging
import pytest
from coauthor.utils.logger import Logger
import os
from coauthor.modules.markdown import process_frontmatter


class MockLogger:
    def info(self, message):
        print(message)

    def debug(self, message):
        print(message)

    def error(self, message):
        print(message)


def test_process_frontmatter():
    """
    Pytest test for process_frontmatter function.
    Tests inclusion of multiple files into the main content.
    """

    with tempfile.TemporaryDirectory() as temp_dir:

        # Create temporary files
        include_file1 = os.path.join(temp_dir, "include1.md")
        with open(include_file1, "w") as f:
            f.write("Content from include1\n\n")

        include_file2 = os.path.join(temp_dir, "include2.md")
        with open(include_file2, "w") as f:
            f.write("Content from include2\n\n")

        frontmatter = f"""---
title: whatever
a:
  b: whatever
include-files: [./include1.md, {include_file2}]
---
"""
        main_file = os.path.join(temp_dir, "main.md")
        with open(main_file, "w") as f:
            f.write(frontmatter + "\nwhatever")

        logger = Logger(__name__, level=logging.DEBUG, log_file=f"{os.getcwd()}/debug.log")

        config = {"current-task": {"path-modify-event": str(main_file), "content": ""}}

        process_frontmatter(config, logger)

        # Verify combined content
        expected_content = f"""{frontmatter}
Content from include1


Content from include2


"""
        logger.debug(f"expected_content: \n{expected_content}")
        logger.debug(f"content: \n{config['current-task']['content']}")
        # TODO remove debug
        # TODO remove debug
        import yaml

        with open("expected_content.txt", "w") as file:
            file.write(expected_content)
        with open("content.txt", "w") as file:
            file.write(config["current-task"]["content"])

        assert config["current-task"]["content"] == expected_content


def test_process_frontmatter_no_include_files():
    """
    Pytest test for process_frontmatter function.
    Tests file with no .
    """

    with tempfile.TemporaryDirectory() as temp_dir:

        frontmatter = """---
title: whatever
a:
  b: whatever
---
"""
        main_file = os.path.join(temp_dir, "main.md")
        with open(main_file, "w") as f:
            f.write(frontmatter + "\nwhatever")

        logger = Logger(__name__, level=logging.DEBUG, log_file=f"{os.getcwd()}/debug.log")

        config = {"current-task": {"path-modify-event": str(main_file), "content": ""}}

        process_frontmatter(config, logger)

        # Verify combined content
        expected_content = f"{frontmatter}\nwhatever"
        logger.debug(f"expected_content: \n{expected_content}")
        logger.debug(f"content: \n{config['current-task']['content']}")
        import yaml

        with open("expected_content.txt", "w") as file:
            file.write(expected_content)
        with open("content.txt", "w") as file:
            file.write(config["current-task"]["content"])

        assert config["current-task"]["content"] == expected_content



import os
import tempfile
import pytest
from unittest import mock
from coauthor.modules.file_processor import pong
from coauthor.modules.file_scanner import scan


# Mock logger for testing
class MockLogger:
    def info(self, message):
        print(message)

    def debug(self, message):
        print(message)


def test_scan_with_pong():
    # Create a temporary directory
    with tempfile.TemporaryDirectory() as temp_dir:
        # Set up a temporary file in the directory
        temp_file_path = os.path.join(temp_dir, "test_file.txt")
        with open(temp_file_path, "w") as temp_file:
            temp_file.write("initial content")

        # Define a mock workflow for testing
        mock_workflows = {
            "workflows": [
                {
                    "name": "pong-workflow",
                    "scan": {"filesystem": {"paths": [temp_dir]}},
                    "tasks": [{"type": "pong", "id": "task1"}],
                }
            ]
        }

        # Create a mock logger
        mock_logger = MockLogger()

        # Call the scan function with the mock workflow and logger
        scan(mock_workflows, mock_logger)

        # Check if the file contents have been changed to 'pong'
        with open(temp_file_path, "r") as temp_file:
            assert temp_file.read() == "pong"


def test_scan_with_invalid_task_type():
    # Create a temporary directory
    with tempfile.TemporaryDirectory() as temp_dir:
        # Set up a temporary file in the directory
        temp_file_path = os.path.join(temp_dir, "test_file.txt")
        with open(temp_file_path, "w") as temp_file:
            temp_file.write("initial content")

        # Define a mock workflow for testing
        mock_workflows = {
            "workflows": [
                {
                    "name": "invalid-task-workflow",
                    "scan": {"filesystem": {"paths": [temp_dir]}},
                    "tasks": [{"type": "nonexistent_task", "id": "task1"}],
                }
            ]
        }

        # Create a mock logger
        mock_logger = MockLogger()

        # Expect the scan function to raise a ValueError due to unknown task type
        with pytest.raises(ValueError, match=r"Unsupported task_type: nonexistent_task"):
            scan(mock_workflows, mock_logger)



import os


def get_recently_modified_file(file_path_inotify, dir_path, logger):
    """
    Retrieves the most recently modified file in a specified directory,
    excluding hidden files. This function is primarily used to address
    an issue with gedit, where the editor uses a temporary hidden file
    that complicates direct file modification detection.

    Args:
        file_path_inotify (str): The file path detected by inotify or similar mechanism.
        dir_path (str): The directory path to search for modified files.
        logger (Logger): A logger instance for logging debug information.

    Returns:
        str: The path to the most recently modified file in the specified directory.
             Returns None if no suitable files are found in the directory.

    Notes:
        - If `file_path_inotify` is found in the directory's non-hidden files,
          it will be returned immediately.
        - For gedit and potentially other applications that use hidden temporary
          files upon saving, this method helps identify the correct file by
          checking the last modified time.
    """
    all_files = list(
        set(
            os.path.join(dir_path, f)
            for f in os.listdir(dir_path)
            if os.path.isfile(os.path.join(dir_path, f)) and not f.startswith(".")
        )
    )
    logger.debug(f"all_files: {', '.join(all_files)}")
    if file_path_inotify in all_files:
        return file_path_inotify
    return max(all_files, key=os.path.getmtime) if all_files else None



import os
import re
import time
import yaml
import json
from threading import Thread
from collections import defaultdict
from jira import JIRA, JIRAError
from coauthor.utils.workflow_utils import (
    get_workflows_that_watch,
)
from coauthor.utils.jira_utils import (
    jira_unanswered_comments,
    execute_jira_query,
    get_jira_connection,
    jira_add_comment,
    get_assigned_tickets,
    jira_update_description_summary,
    jira_assign_to_creator,
)
from coauthor.modules.ai import process_file_with_openai_agent


def parse_ai_response(response, logger):
    log_message = re.sub(r"[\r\n]+", " ", str(response))[:100]
    logger.info(f"AI response → {log_message}")
    try:
        response_json = json.loads(response)
    except (json.JSONDecodeError, TypeError) as exception_error:
        logger.error(f"Error parsing JSON response: {exception_error}")
        response_json = response if isinstance(response, dict) else None
    return response_json


task_type_functions = {
    "ai": process_file_with_openai_agent,
}


def watch(config, logger):
    logger.info("Watching Jira workflows")
    # watch_jira(config, logger)
    new_thread = Thread(target=watch_jira, args=(config, logger))
    new_thread.start()


def watch_jira(config, logger):
    workflows_that_watch_jira = get_workflows_that_watch(config, logger, "jira")
    while True:
        for workflow in workflows_that_watch_jira:
            config["current-workflow"] = workflow
            jira_config = workflow["watch"]["jira"]
            sleep_time = jira_config.get("sleep", 15)
            logger.debug(f"Starting thread for workflow {workflow['name']}, sleep {sleep_time} seconds")
            process_jira_comments(config, logger)
            process_jira_tickets(config, logger)
            time.sleep(sleep_time)


def process_jira_comments(config, logger):
    workflow = config["current-workflow"]
    jira_config = workflow["watch"]["jira"]
    jira_url = jira_config.get("url") or os.environ.get("COAUTHOR_JIRA_URL")
    logger.debug(f"Checking Jira in workflow {workflow['name']}")
    jira_query = jira_config.get("query", "updated >= -0.3h OR created >= -0.3h")
    logger.debug(f"jira_query: {jira_query}")
    if not jira_url or not jira_query:
        logger.error(f"JIRA configuration missing. Skipping. jira_config: {jira_config}")
        return
    jira_instance = get_jira_connection(jira_url, logger)
    config["current-jira-instance"] = jira_instance
    if not jira_instance:
        return
    all_tickets = execute_jira_query(jira_instance, jira_query, logger)
    if not all_tickets:
        logger.debug("No Jira tickets found")
        return
    tickets = jira_unanswered_comments(config, logger, all_tickets)
    if not tickets:
        logger.debug("No Jira tickets with a content match found")
        return
    ticket_list = [f"{ticket['ticket'].key} - {ticket['ticket'].fields.summary}" for ticket in tickets]
    logger.info(f"Found tickets with unanswered comments: {', '.join(ticket_list)}")
    for ticket in tickets:
        config["current-ticket"] = ticket["ticket"]
        config["current-comment"] = ticket["comment"]
        get_related_tickets(config, logger, ticket["ticket"])
        handle_jira_comments(config, logger)


def handle_jira_comments(config, logger):
    comment = config["current-comment"]
    ticket = config["current-ticket"]
    workflow = config["current-workflow"]
    logger.info(f"Processing ticket {ticket.key}")
    for task in workflow["tasks"]:
        task["current-ticket"] = ticket
        task["related-tickets"] = config["related-tickets"]
        task["current-comment"] = comment
        logger.debug(f"task: {task}")
        if task["type"] in task_type_functions:
            config["current-task"] = task
            task["json"] = True
            logger.debug(f"Workflow: {workflow['name']}, Task: {task['id']} → {ticket.key}")
            config["current-task"] = task
            task_type_functions[task["type"]](config, logger)
            response = task["response"]
            response_json = parse_ai_response(response, logger)
            if response_json and isinstance(response_json, dict):
                if "comment" in response_json:
                    comment_text = response_json["comment"]
                    labels = response_json.get("labels", None)
                    labels_remove = response_json.get("labels_remove", None)
                    jira_add_comment(config, logger, ticket, comment_text, labels, labels_remove)
                else:
                    logger.error("Response JSON missing 'comment' key")
            else:
                logger.error("Invalid response for comment processing")
        else:
            raise ValueError(f'Unsupported task_type: {task["type"]}')


def get_related_tickets(config, logger, ticket):
    jira_instance = config["current-jira-instance"]
    related = {}

    if hasattr(ticket.fields, "parent") and ticket.fields.parent:
        try:
            related["parent"] = jira_instance.issue(ticket.fields.parent.key, fields="summary,description")
        except JIRAError as jira_error:
            logger.error(f"Failed to fetch parent for {ticket.key}: {jira_error}")

    # See Jira Settings > Issues > Custom fields
    possible_epic_fields = ["customfield_10011"]
    epic_key = None
    for field in possible_epic_fields:
        epic_key = getattr(ticket.fields, field, None)
        if epic_key:
            break
    if epic_key:
        try:
            related["epic"] = jira_instance.issue(epic_key, fields="summary,description")
        except JIRAError as jira_error:
            logger.error(f"Failed to fetch epic {epic_key} for {ticket.key}: {jira_error}")

    # Fetch linked issues
    linked_issues = []
    if hasattr(ticket.fields, "issuelinks") and ticket.fields.issuelinks:
        for link in ticket.fields.issuelinks:
            if hasattr(link, "outwardIssue") and link.outwardIssue:
                try:
                    linked_ticket = jira_instance.issue(link.outwardIssue.key, fields="summary,description")
                    linked_issues.append({"type": link.type.outward.lower(), "ticket": linked_ticket})
                except JIRAError as jira_error:
                    logger.error(f"Failed to fetch outward link {link.outwardIssue.key} for {ticket.key}: {jira_error}")
            elif hasattr(link, "inwardIssue") and link.inwardIssue:
                try:
                    linked_ticket = jira_instance.issue(link.inwardIssue.key, fields="summary,description")
                    linked_issues.append({"type": link.type.inward.lower(), "ticket": linked_ticket})
                except JIRAError as jira_error:
                    logger.error(f"Failed to fetch inward link {link.inwardIssue.key} for {ticket.key}: {jira_error}")
    related["linked"] = linked_issues

    # Now group dynamically
    grouped = defaultdict(list)
    for link in related["linked"]:
        grouped[link["type"]].append({"ticket": link["ticket"]})
    related["grouped_linked"] = dict(grouped)

    config["related-tickets"] = related


def process_jira_tickets(config, logger):
    workflow = config["current-workflow"]
    jira_config = workflow["watch"]["jira"]
    jira_url = jira_config.get("url") or os.environ.get("COAUTHOR_JIRA_URL")
    logger.debug(f"Checking assigned Jira tickets in workflow {workflow['name']}")
    if not jira_url:
        logger.error(f"JIRA URL missing. Skipping. jira_config: {jira_config}")
        return
    if "current-jira-instance" in config:
        jira_instance = config["current-jira-instance"]
    else:
        jira_instance = get_jira_connection(jira_url, logger)
        config["current-jira-instance"] = jira_instance
    if not jira_instance:
        return
    tickets = get_assigned_tickets(jira_instance, logger)
    if not tickets:
        logger.debug("No assigned Jira tickets found")
        return
    ticket_list = [f"{ticket.key} - {ticket.fields.summary}" for ticket in tickets]
    logger.info(f"Found assigned tickets: {', '.join(ticket_list)}")
    for ticket in tickets:
        config["current-ticket"] = ticket
        get_related_tickets(config, logger, ticket)
        handle_jira_tickets(config, logger)


def handle_jira_tickets(config, logger):
    ticket = config["current-ticket"]
    workflow = config["current-workflow"]
    logger.info(f"Processing assigned ticket {ticket.key}")
    for task in workflow["tasks"]:
        task["current-ticket"] = ticket
        task["related-tickets"] = config["related-tickets"]
        if "current-comment" in task:
            del task["current-comment"]
        logger.debug(f"task: {task}")
        if task["type"] in task_type_functions:
            logger.debug(f"Workflow: {workflow['name']}, Task: {task['id']} → {ticket.key}")
            config["current-task"] = task
            task["json"] = True
            task_type_functions[task["type"]](config, logger)
            response = task["response"]
            response_json = parse_ai_response(response, logger)
            if response_json and isinstance(response_json, dict):
                if "summary" in response_json and "description" in response_json:
                    summary = response_json["summary"]
                    description = response_json["description"]
                    labels = response_json.get("labels", None)
                    labels_remove = response_json.get("labels_remove", None)
                    story_points = response_json.get("story_points", None)
                    jira_update_description_summary(
                        config, logger, ticket, summary, description, story_points, labels, labels_remove
                    )
                else:
                    logger.error("Response JSON missing 'summary' or 'description' key")
            else:
                logger.error("Invalid response for ticket update")
            try:
                jira_assign_to_creator(logger, ticket)
            except Exception as exception_error:
                logger.error(f"Failed to assign ticket {ticket.key} to creator: {exception_error}")
                raise ValueError(f"Assignment failed for {ticket.key}: {exception_error}") from exception_error
        else:
            raise ValueError(f'Unsupported task_type: {task["type"]}')



import os
import time
from inotify_simple import INotify, flags

from coauthor.utils.workflow_utils import (
    get_workflows_that_watch,
    get_all_watch_directories_from_workflows,
)
from coauthor.utils.match_utils import file_path_match, file_content_match
from coauthor.utils.inotify_utils import get_recently_modified_file
from coauthor.modules.ai import process_file_with_openai_agent
from coauthor.modules.file_processor import regex_replace_in_file, pong, include_file
from coauthor.modules.workflow_tasks import write_file, read_file
from coauthor.modules.plantuml import process_puml_file
from coauthor.modules.markdown import process_frontmatter
from coauthor.modules.web_tasks import replace_redirecting_links
from coauthor.modules.youtube import get_youtube_video_info


task_type_functions = {
    "ai": process_file_with_openai_agent,
    "regex_replace_in_file": regex_replace_in_file,
    "pong": pong,
    "write_file": write_file,
    "read_file": read_file,
    "plantuml": process_puml_file,
    "include-files": process_frontmatter,
    "include-file": include_file,  # TODO better naming
    "replace_redirecting_links": replace_redirecting_links,
    "youtube": get_youtube_video_info,
}

# Initialize a dictionary to store the last modification time of each file based
# on workflow and task
last_modification_times = {}


def add_watch_recursive(proj, inotify, directory):
    """Recursively add watches on all subdirectories,
    ignoring certain directories.

    Args:
        proj (dict): The project configuration.
        inotify (INotify): The INotify instance to add watches.
        directory (str): The root directory to begin adding watches.

    Returns:
        dict: A dictionary mapping watch descriptors to paths.
    """
    ignore_folders = proj.get("file-watcher", {}).get("ignore-folders", [])
    wd_to_path = {}
    for root, dirs, _ in os.walk(directory):
        # Ignore specific directories like __pycache__ and .obsidian
        dirs[:] = [d for d in dirs if d not in ignore_folders]
        if any(ignore in root for ignore in ignore_folders):
            continue
        watch_descriptor = inotify.add_watch(root, flags.CREATE | flags.MODIFY)
        wd_to_path[watch_descriptor] = root
    return wd_to_path


def watch(config, logger):
    """Start watching directories specified in the config and handle events.

    Args:
        config (dict): Configuration settings for the watcher.
        logger (Logger): A logger object for logging messages.
    """
    watch_directory(config, logger)


def get_project_for_path(file_path, path_to_project):
    """Find the project configuration for a given file path using longest prefix match.

    Args:
        file_path (str): The path of the file.
        path_to_project (dict): Mapping of project root paths to project configs.

    Returns:
        dict: The project configuration, or None if no match.
    """
    matches = []
    abs_path = os.path.abspath(os.path.expanduser(file_path))
    for root, proj in path_to_project.items():
        if abs_path.startswith(root):
            matches.append((len(root), proj))
    if not matches:
        return None
    matches.sort(reverse=True)
    return matches[0][1]


def handle_inotify_event(event, wd_to_path, inotify, path_to_project, logger):
    """Handle inotify events by accurately determining the file changed.

    For MODIFY events, due to some editors' behavior of replacing files rather than directly modifying them,
    this function identifies the most recently updated file in the event's directory as the file affected by the event.

    Args:
        event (Event): The inotify event object.
        wd_to_path (dict): A mapping of watch descriptors to directory paths.
        inotify (INotify): The INotify instance handling the events.
        path_to_project (dict): Mapping of project roots to configs.
        logger (Logger): A logger object for logging messages.

    Returns:
        bool: True if a stop file was detected and processed, otherwise False.
    """
    directory = wd_to_path.get(event.wd, ".")
    logger.debug(
        f"Inotify Event: directory={directory}, event_mask={event.mask}, flags.from_mask={flags.from_mask(event.mask)}"
    )

    proj = get_project_for_path(directory, path_to_project)
    if not proj:
        logger.warning(f"No project found for directory {directory}")
        return False

    file_path_inotify = os.path.join(directory, event.name)
    file_path = get_recently_modified_file(file_path_inotify, directory, logger)

    # Check if it is a directory and add watch if new directory is created
    if flags.CREATE in flags.from_mask(event.mask) and os.path.isdir(file_path_inotify):
        logger.info(f"Watching new directory: {file_path_inotify}")
        new_wd_to_path = add_watch_recursive(proj, inotify, file_path_inotify)
        wd_to_path.update(new_wd_to_path)

    if file_path_inotify != file_path:
        logger.warning(f"file_path_inotify: {file_path_inotify} is not equal to file_path: {file_path}")
        logger.debug(" this can occur depending on how editors write changes to files")
        logger.debug(" For example Gedit uses a temporary file .goutputstream-G1SHX2")
        file_path_selected = file_path
        time.sleep(2)  # allow Gedit some time to finish updating file
    else:
        file_path_selected = file_path_inotify

    if is_binary_file(file_path_selected, logger):
        logger.debug(f"Ignoring binary file: {file_path_selected}")
        return False

    ignore_extensions = proj.get("watch-ignore-file-extensions", [])
    if flags.MODIFY in flags.from_mask(event.mask) and file_path_selected:
        file_extension = os.path.splitext(file_path_selected)[1]
        if file_extension not in ignore_extensions:
            logger.info(f"MODIFY event: file_path_inotify: {file_path_selected}")
            handle_workflow_file_modify_event(file_path_selected, proj, logger)
        else:
            logger.debug(f"Ignore MODIFY event {file_path_selected} because of file extension {file_extension}")

    if os.path.basename(file_path_selected) == "stop":
        logger.info("Stop file found!")
        os.remove(file_path_selected)
        return True

    return False


def is_binary_file(filepath, logger):
    """Check if a file is binary by reading a portion of its content.

    Args:
        filepath (str): The path to the file.
        logger (Logger): A logger object for logging errors.

    Returns:
        bool: True if the file is binary, False otherwise.
    """
    if not filepath or os.path.isdir(filepath):
        logger.debug(f"Skipping non-file or None path: {filepath}")
        return False
    try:
        with open(filepath, "rb") as file:
            # Read part of the file to determine if it's binary
            chunk = file.read(1024)
            if b"\0" in chunk:
                return True
    except OSError as os_error:
        # Log any exception that might occur while reading the file
        logger.error(f"Error reading file {filepath}: {os_error}")
    return False


def watch_directory(config, logger):
    """Continuously watch specified directories for changes and handle events.

    Args:
        config (dict): Configuration settings specifying directories to watch.
        logger (Logger): A logger object for logging messages.
    """

    project_list = []
    if "projects" in config:
        project_list = config["projects"] + [config]
    else:
        project_list = [config]

    all_watch_directories = []
    path_to_project = {}

    inotify = INotify()
    wd_to_path = {}

    for proj in project_list:
        proj_path = os.path.expanduser(proj.get("path", os.getcwd()))
        abs_proj_path = os.path.abspath(proj_path)
        path_to_project[abs_proj_path] = proj

        if "workflows" in proj:
            proj_dirs = get_all_watch_directories_from_workflows(proj, logger)
            abs_dirs = [
                (
                    os.path.abspath(os.path.join(proj_path, os.path.expanduser(d)))
                    if not os.path.isabs(os.path.expanduser(d))
                    else os.path.expanduser(d)
                )
                for d in proj_dirs
            ]
            all_watch_directories.extend(abs_dirs)

            for abs_dir in abs_dirs:
                wd_to_path.update(add_watch_recursive(proj, inotify, abs_dir))
        else:
            logger.debug(f"Skipping project {proj_path} as it has no workflows")

    # Deduplicate watch directories if necessary
    watch_directories = list(set(all_watch_directories))
    logger.info(f"Watching directories recursively: {', '.join(watch_directories)}")

    try:
        while True:
            for event in inotify.read():
                if handle_inotify_event(event, wd_to_path, inotify, path_to_project, logger):
                    return
    except KeyboardInterrupt:
        logger.info("Stopping directory watch")
    finally:
        for watch_descriptor in wd_to_path:
            inotify.rm_watch(watch_descriptor)


def handle_workflow_file_modify_event(path, proj, logger):
    """Handle file modification events by checking workflows and invoking tasks.

    Args:
        path (str): The path of the modified file triggering the event.
        proj (dict): Project configuration settings specifying workflows and tasks.
        logger (Logger): A logger object for logging messages.
    """
    workflows_that_watch = get_workflows_that_watch(proj, logger)
    logger.debug(f"get_workflows_that_watch: workflows_that_watch: {workflows_that_watch}")
    for workflow in workflows_that_watch:
        proj["current-workflow"] = workflow
        logger.debug(f"workflow: {workflow}")
        for task in workflow["tasks"]:
            proj["current-task"] = task
            task["path-modify-event"] = path
            path_match = file_path_match(proj, logger)
            if "content_patterns" in workflow:
                content_match = file_content_match(proj, logger)
            else:
                logger.debug('Workflow has no "content_patterns", so "content_match" is True')
                content_match = True
            if not path_match or not content_match:
                logger.debug(f"No path or content match on {path}")
                continue
            logger.info(f"Path and content match on {path}")
            if task["type"] in task_type_functions:
                logger.info(f"Workflow: {workflow['name']}, Task: {task['id']} → {path}")
                if not ignore_rapid_modify_event(path, proj, logger):
                    task_type_functions[task["type"]](proj, logger)
            else:
                raise ValueError(f'Unsupported task_type: {task["type"]}')


def ignore_rapid_modify_event(path, proj, logger):
    """Determine whether to ignore a rapid succession of modify events.

    Args:
        path (str): The path of the file associated with the event.
        proj (dict): Project configuration settings including workflow and task info.
        logger (Logger): A logger object for logging messages.

    Returns:
        bool: True if the event is ignored, False otherwise.
    """
    current_time = time.time()
    workflow = proj["current-workflow"]
    task = proj["current-task"]
    key = (path, workflow["name"], task["id"])
    last_time = last_modification_times.get(key, 0)

    modify_event_limit = 3
    if "modify_event_limit" in task:
        modify_event_limit = task["modify_event_limit"]

    if current_time - last_time < modify_event_limit:
        logger.info(f"  Ignoring rapid MODIFY event for (modify_event_limit: {modify_event_limit}).")
        return True
    last_modification_times[key] = current_time
    logger.debug(f"  NOT ignoring rapid MODIFY event for (modify_event_limit: {modify_event_limit}).")
    return False



import yaml


def parse_frontmatter(content):
    """
    Parse frontmatter from markdown content without using frontmatter library
    because it depends on older pyyaml==5.3.1.

    Returns a dictionary with 'data' (parsed frontmatter) and 'content'.
    """
    if "---" not in content:
        return {"frontmatter": {}, "content": content}

    frontmatter_end = content.find("---", 3)
    if frontmatter_end == -1:
        frontmatter_end = 3

    frontmatter_str = content[3:frontmatter_end].strip()
    frontmatter_data = yaml.safe_load(frontmatter_str)

    if frontmatter_end + 1 < len(content):
        content_str = content[frontmatter_end + 1 :].lstrip()
    else:
        content_str = ""

    return {"frontmatter": frontmatter_data, "content": content_str, "frontmatter_str": frontmatter_str}


def get_frontmatter(path):
    """
    Reads a Markdown file and parses its frontmatter.

    :param path: Path to the Markdown file.
    :return: Tuple containing the frontmatter dictionary, frontmatter string, and original content.
    """
    with open(path, "r", encoding="utf-8") as file:
        content = file.read()
    parsed = parse_frontmatter(content)
    frontmatter = parsed["frontmatter"]
    frontmatter_str = parsed["frontmatter_str"]
    return frontmatter, frontmatter_str, content


def get_frontmatter_attribute(path, attr_name):
    """
    Retrieves a specific attribute from the frontmatter of a Markdown file.

    :param path: Path to the Markdown file.
    :param attr_name: The name of the attribute to retrieve.
    :return: The value of the specified attribute, or None if not found.
    """
    frontmatter, _, _ = get_frontmatter(path)
    return (frontmatter or {}).get(attr_name)


def get_frontmatter_nested_value(path, frontmatter_context_path):
    """
    Retrieves a nested value from the frontmatter of a Markdown file using a path.

    :param path: Path to the Markdown file.
    :param frontmatter_context_path: The nested path to the attribute (e.g., "key1/key2").
    :return: The value as a list if it is a list, otherwise an empty list.
    """
    frontmatter, _, _ = get_frontmatter(path)
    keys = frontmatter_context_path.replace("/", ".").split(".")
    value = frontmatter
    for key in keys:
        value = value.get(key, {})
    return value if isinstance(value, list) else []



#   ---------------------------------------------------------------------------------
#   Copyright (c) Microsoft Corporation. All rights reserved.
#   Licensed under the MIT License. See LICENSE in project root for information.
#   ---------------------------------------------------------------------------------
"""
This is a configuration file for pytest containing customizations and fixtures.

In VSCode, Code Coverage is recorded in config.xml. Delete this file to reset reporting.
"""

from __future__ import annotations

from typing import List

import pytest
from _pytest.nodes import Item
import os
import time

pytest_plugins = ["helpers_namespace"]


def pytest_collection_modifyitems(items: list[Item]):
    for item in items:
        if "spark" in item.nodeid:
            item.add_marker(pytest.mark.spark)
        elif "_int_" in item.nodeid:
            item.add_marker(pytest.mark.integration)


@pytest.fixture
def unit_test_mocks(monkeypatch: None):
    """Include Mocks here to execute all commands offline and fast."""
    pass


# @pytest.fixture
# def load_test_data():
#     data_file = os.path.join(os.path.dirname(__file__), 'data', 'test_data.json')
#     with open(data_file, 'r') as f:
#         return json.load(f)


@pytest.helpers.register
def wait_file_content(file_path: str, expected_content: str, retries: int = 5, delay: int = 1):
    """
    Wait for a file's content to match the expected content.

    Args:
        file_path (str): Path to the file whose content needs to be checked.
        expected_content (str): The content that we expect the file to have.
        retries (int): Number of retries before giving up. Defaults to 5.
        delay (int): Delay in seconds between each retry. Defaults to 1 second.

    Returns:
        str: The actual content of the file after all retries.
    """
    actual_content = ""
    for _ in range(retries):
        with open(file_path, "r") as file:
            actual_content = file.read()
            if actual_content == expected_content:
                return actual_content
        time.sleep(delay)
    return actual_content



from jira import JIRA, JIRAError
from coauthor.utils.notify import notification
import os

import re


def get_jira_connection(jira_url, logger):
    jira_username = os.getenv("COAUTHOR_JIRA_USERNAME")
    jira_password = os.getenv("COAUTHOR_JIRA_PASSWORD")

    try:
        jira_instance = JIRA(jira_url, basic_auth=(jira_username, jira_password))
        logger.debug(f"Connected to JIRA at {jira_url}")
        return jira_instance
    except JIRAError as jira_error:
        logger.error(f"Failed to connect to JIRA at {jira_url}: {jira_error}")
        return


def execute_jira_query(jira_instance, query, logger):
    try:
        issues = jira_instance.search_issues(query)
        logger.debug(f"Found {len(issues)} issues for query: {query}")
        return issues
    except (JIRAError, ValueError) as exception_error:
        logger.error(f"Error executing JIRA query: {exception_error}")
        return


def jira_ticket_unanswered_comment(ticket, content_patterns, jira_username, logger):
    comments = ticket.fields.comment.comments
    if not comments:
        return None

    last_match_index = -1
    for i, comment in enumerate(comments):
        body = comment.body
        log_message = re.sub(r"[\r\n]+", " ", body)[:100]
        logger.debug(f"{comment.author.name} → {log_message}")
        if not comment.author.name == jira_username:  # don't answer coauthor own comments
            for pattern in content_patterns:
                if re.search(pattern, body):
                    log_message = re.sub(r"[\r\n]+", " ", body)[:100]
                    logger.debug(f'{ticket.key} matches pattern "{pattern}": {log_message}')
                    last_match_index = i
                    break

    if last_match_index == -1:
        return None

    # Check if there's any comment after the last matched one by jira_username
    answered = False
    for i in range(last_match_index + 1, len(comments)):
        comment = comments[i]
        if comment.author.name == jira_username:
            answered = True
            break

    if answered:
        return None

    return comments[last_match_index]


def jira_unanswered_comments(config, logger, tickets):
    """Check Jira tickets for unanswered content matches against defined patterns in comments."""
    jira_username = os.getenv("COAUTHOR_JIRA_USERNAME")
    workflow = config.get("current-workflow", {})
    content_patterns = workflow.get("content_patterns")
    if not content_patterns:
        logger.debug("No content patterns defined in workflow.")
        return []

    matching_tickets = []
    for ticket in tickets:
        unanswered_comment = jira_ticket_unanswered_comment(ticket, content_patterns, jira_username, logger)
        if unanswered_comment is not None:
            matching_tickets.append({"ticket": ticket, "comment": unanswered_comment})

    return matching_tickets


def get_updated_labels(current_labels, labels=None, labels_remove=None, logger=None, ticket_key=None):
    current_labels = current_labels[:]  # copy
    changed = False

    if labels_remove:
        removed_labels = [l for l in labels_remove if l in current_labels]
        if removed_labels:
            current_labels = [l for l in current_labels if l not in labels_remove]
            logger.info(f"{ticket_key} removed labels → {', '.join(removed_labels)}")
            changed = True

    if labels:
        added_labels = [l for l in labels if l not in current_labels]
        if added_labels:
            current_labels.extend(added_labels)
            logger.info(f"{ticket_key} labels → {', '.join(added_labels)}")
            changed = True

    return current_labels, changed


def jira_add_comment(config, logger, ticket, content, labels=None, labels_remove=None):
    jira_instance = config["current-jira-instance"]
    try:
        comment = jira_instance.add_comment(ticket, content)
        logger.debug(f"Added comment to {ticket.key} → {comment.id}")

        updated_fields = []

        current_labels = ticket.fields.labels or []
        new_labels, changed = get_updated_labels(current_labels, labels, labels_remove, logger, ticket.key)
        if changed:
            ticket.update(fields={"labels": new_labels})
            updated_fields.append("labels")

        args = config.get("args", None)
        if args and args.notify:
            notification(f"Coauthor updated {ticket.key}", "Coauthor added a comment to the Jira ticket")
        if updated_fields:
            logger.info(f"Updated {', '.join(updated_fields)} for {ticket.key}")
        return comment
    except JIRAError as error:
        logger.error(f"Failed to add comment to {ticket.key}: {error}")
        return None


def get_assigned_tickets(jira_instance, logger):
    jira_username = os.getenv("COAUTHOR_JIRA_USERNAME")
    if not jira_username:
        logger.error("Jira username/COAUTHOR_JIRA_USERNAME not set")
        return []
    query = f'assignee = "{jira_username}"'
    return execute_jira_query(jira_instance, query, logger) or []


def jira_update_description_summary(
    config, logger, ticket, summary, description, story_points=None, labels=None, labels_remove=None
):
    try:
        update_kwargs = {"summary": summary, "description": description}
        custom_fields = {}
        updated_fields = ["summary", "description"]

        if story_points is not None and ticket.fields.issuetype.name == "Story":
            story_point_field = "customfield_10007"  # TODO configure
            current_sp = getattr(ticket.fields, story_point_field, None)
            if current_sp != story_points:
                logger.info(f"{ticket.key} story_points → {story_points}")
                custom_fields[story_point_field] = story_points

        current_labels = ticket.fields.labels or []
        new_labels, changed = get_updated_labels(current_labels, labels, labels_remove, logger, ticket.key)
        if changed:
            custom_fields["labels"] = new_labels
            updated_fields.append("labels")

        ticket.update(**update_kwargs, fields=custom_fields)
        args = config.get("args", None)
        if args and args.notify:
            notification(f"Coauthor updated {ticket.key}", "Coauthor updated the Jira ticket")
        logger.info(f"Updated {', '.join(updated_fields)} for {ticket.key}")
        return True
    except JIRAError as error:
        logger.error(f"Failed to update summary and description for {ticket.key}: {error}")
        return False


def jira_assign_to_creator(logger, ticket):
    reporter = ticket.fields.reporter.name
    try:
        ticket.update(assignee={"name": reporter})
        logger.info(f"Assigned {ticket.key} to creator {reporter}")
        return True
    except JIRAError as error:
        logger.error(f"Failed to assign {ticket.key} to creator: {error}")
        return False



import os
import pytest
from unittest import mock
from coauthor.utils.inotify_utils import get_recently_modified_file
from logging import Logger


def test_get_recently_modified_file_with_inotify_file():
    # Setup
    logger = mock.Mock(Logger)
    dir_path = "/mocked/path"
    inotify_file = "/mocked/path/file.txt"

    mocked_files = {"file.txt": 1, "another_file.txt": 0}  # File has been modified recently

    def mock_isfile(path):
        return path.split("/")[-1] in mocked_files

    def mock_listdir(path):
        return list(mocked_files.keys())

    def mock_getmtime(path):
        filename = path.split("/")[-1]
        return mocked_files[filename]

    with mock.patch("os.listdir", mock_listdir), mock.patch("os.path.isfile", mock_isfile), mock.patch(
        "os.path.getmtime", mock_getmtime
    ):
        # Execute
        result = get_recently_modified_file(inotify_file, dir_path, logger)

    # Verify
    assert result == inotify_file
    logger.debug.assert_called()


def test_get_recently_modified_file_without_inotify_file():
    # Setup
    logger = mock.Mock(Logger)
    dir_path = "/mocked/path"
    inotify_file = "/mocked/path/nonexistent.txt"

    mocked_files = {"file.txt": 1, "another_file.txt": 2}  # This one is the most recently modified

    def mock_isfile(path):
        return path.split("/")[-1] in mocked_files

    def mock_listdir(path):
        return list(mocked_files.keys())

    def mock_getmtime(path):
        filename = path.split("/")[-1]
        return mocked_files[filename]

    with mock.patch("os.listdir", mock_listdir), mock.patch("os.path.isfile", mock_isfile), mock.patch(
        "os.path.getmtime", mock_getmtime
    ):
        # Execute
        result = get_recently_modified_file(inotify_file, dir_path, logger)

    # Verify
    assert result == "/mocked/path/another_file.txt"
    logger.debug.assert_called()


def test_get_recently_modified_file_empty_directory():
    # Setup
    logger = mock.Mock(Logger)
    dir_path = "/mocked/path"
    inotify_file = "/mocked/path/nonexistent.txt"

    def mock_isfile(path):
        return False

    def mock_listdir(path):
        return []

    with mock.patch("os.listdir", mock_listdir), mock.patch("os.path.isfile", mock_isfile):
        # Execute
        result = get_recently_modified_file(inotify_file, dir_path, logger)

    # Verify
    assert result is None
    logger.debug.assert_called()



def get_all_watch_directories_from_workflows(config, logger):
    return get_all_directories_from_workflows(config, logger, "watch")


def get_all_scan_directories_from_workflows(config, logger):
    return get_all_directories_from_workflows(config, logger, "scan")


def get_all_directories_from_workflows(config, logger, watch_or_scan_key):
    dirs = set()  # Use a set to ensure that directories are unique
    for workflow in config["workflows"]:
        if "filesystem" in workflow[watch_or_scan_key]:
            dirs.update(workflow[watch_or_scan_key]["filesystem"]["paths"])
    return list(dirs)  # Convert the set back to a list for the return value


def get_workflows_that_watch(config, logger, watch_filesystem_or="filesystem"):
    wtw = []
    workflows = config.get("workflows", [])  # Default naar een lege lijst
    for workflow in workflows:
        if "watch" in workflow:
            if watch_filesystem_or in workflow["watch"]:
                wtw.append(workflow)
    logger.debug(f"get_workflows_that_watch: workflows_that_watch: {wtw}")
    return wtw


def get_workflows_that_scan(config, logger):
    workflows_that_scan = []
    workflows = config.get("workflows", [])

    for workflow in workflows:
        if "scan" in workflow:
            workflows_that_scan.append(workflow)
    logger.debug(f"get_workflows_that_scan: workflows_that_scan: {workflows_that_scan}")
    return workflows_that_scan



import os
import tempfile
import logging
import pytest
from coauthor.utils.logger import Logger
from unittest.mock import patch, mock_open, MagicMock
from coauthor.utils.config import get_config
from coauthor.modules.ai import process_file_with_openai_agent, process_with_openai_agent

import yaml
import jinja2
import openai
from tempfile import TemporaryDirectory


def get_pong_config():
    config_path = os.path.join(os.path.dirname(__file__), "data", "coauthor_task_pong.yml")
    with open(config_path, "r") as file:
        config = yaml.safe_load(file)
        config["current-workflow"] = config["workflows"][0]
        config["current-task"] = config["current-workflow"]["tasks"][0]
    return config


def test_process_file_with_openai_agent():
    """
    Test the `process_file_with_openai_agent` function by creating a temporary
    file with a specific pattern, processing it, and verifying that the content
    changes as expected. This test also verifies that task dict is updated by
    process_file_with_openai_agent with the path of the file.

    The function should transform the content from a placeholder string (ai-test: ping)
    to 'pong' using the OpenAI agent.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        config = get_pong_config()
        config["jinja"] = {"search_path": temp_dir}
        logger = Logger(__name__, level=logging.DEBUG, log_file=f"{os.getcwd()}/debug.log")
        task = config["current-task"]
        logger.debug(f"task:{task}")
        file_path = os.path.join(temp_dir, "whatever.md")

        with open(file_path, "w") as f:
            f.write("(ai-test: ping)")

        process_file_with_openai_agent(config, logger)
        assert task["response"] == "pong"


def test_process_with_openai_agent():
    logger = Logger(__name__, level=logging.INFO)
    with tempfile.TemporaryDirectory() as temp_dir:
        config = get_pong_config()
        config["jinja"] = {"search_path": temp_dir}
        task = config["current-task"]
        task["user"] = "(ai-test: ping)"
        response = process_with_openai_agent(config, logger)
        assert response == "pong"


def test_process_with_openai_agent_write_response_to_yaml():
    logger = Logger(__name__, level=logging.INFO)
    with tempfile.TemporaryDirectory() as temp_dir:
        config = get_pong_config()
        config["jinja"] = {"search_path": temp_dir}
        task = config["current-task"]
        task["user"] = "(ai-test: ping)"

        coauthor_ai_log_dir = tempfile.mkdtemp()  # Create a temporary directory to act as COAUTHOR_AI_LOG_DIR
        os.environ["COAUTHOR_AI_LOG_DIR"] = coauthor_ai_log_dir  # Set environment variable

        response = process_with_openai_agent(config, logger)

        assert response == "pong"

        # Assert that the ai-log-file exists and is updated in the task
        assert "ai-log-file" in task
        assert os.path.exists(task["ai-log-file"])


def test_process_with_wrong_openai_agent_config_raises_exception():
    # Define the sample content and agent
    logger = Logger(__name__, level=logging.INFO)
    with tempfile.TemporaryDirectory() as temp_dir:
        config = get_pong_config()
        config["jinja"] = {"search_path": temp_dir}
        task = config["current-task"]
        task["user"] = "whatever"
        workflow = config["current-workflow"]
        del config["agent"]["api_key_var"]
        del config["agent"]["api_url_var"]
        config["agent"]["api_key_var"] = "NOKEYATALL"
        config["agent"]["api_url"] = "http://9554090.com"

        with pytest.raises(openai.APIConnectionError) as excinfo:
            process_with_openai_agent(config, logger)

        assert "Connection error" in str(excinfo.value)



import os
import tempfile
from coauthor.utils.jinja import render_template
import logging
from coauthor.utils.logger import Logger
import yaml
import subprocess
from coauthor.utils.jinja_filters import (
    select_task,
    get_task_attribute,
    TaskNotFoundError,
    AttributeNotFoundError,
    dirname,
    basename,
)


def get_config(jinja_search_path, some_file_to_include):
    return yaml.safe_load(
        f"""
---
jinja:
  search_path: {jinja_search_path}
workflows:
  - name: jinja-filters-test
    tasks:
      - id: test_jinja_filter_include_file_content
        path-modify-event: {some_file_to_include}
"""
    )


def get_config2():
    return yaml.safe_load(
        f"""
---
workflows:
  - name: whatever-workflow
    tasks:
      - id: task1
        name: First Task
        priority: High
      - id: task2
        name: Second Task
        priority: Low
"""
    )


def test_jinja_filter_include_file_content():
    """
    This test verifies we include files in content we render using Jinja
    template engine using Jinja filter include_file_content
    """
    logger = Logger(__name__, level=logging.DEBUG, log_file=f"{os.getcwd()}/debug.log")
    with tempfile.TemporaryDirectory() as temp_dir:
        some_file_to_include = os.path.join(temp_dir, "some_file_to_include.md")
        with open(some_file_to_include, "w") as f:
            f.write("some_content_to_include")
        path_user_template = os.path.join(temp_dir, "templates", "user.md")
        path_templates_dir = os.path.dirname(path_user_template)
        os.makedirs(path_templates_dir)
        with open(path_user_template, "w") as f:
            f.write('{{ config["current-task"]["path-modify-event"] | include_file_content }}')
        config = get_config(path_templates_dir, some_file_to_include)
        config["current-workflow"] = config["workflows"][0]
        task = config["current-workflow"]["tasks"][0]
        config["current-task"] = task
        template = os.path.basename(path_user_template)
        content = render_template(task, template, config, logger)
        assert content == "some_content_to_include"


def test_jinja_filter_get_git_diff():
    """
    This test setups a local git repository, adds a file, makes a local change,
    and then verifies the rendering of git diff content using the Jinja filter 'get_git_diff'.
    Also checks that files not added to git do not produce a diff
    and committed files with no changes return none.
    """

    logger = Logger(__name__, level=logging.DEBUG, log_file=f"{os.getcwd()}/debug.log")

    with tempfile.TemporaryDirectory() as temp_dir:
        # Initialize a git repository
        subprocess.run(["git", "init"], cwd=temp_dir, check=True)

        # Add and commit a file with content
        diff_file_path = os.path.join(temp_dir, "diff_file.txt")
        with open(diff_file_path, "w") as f:
            f.write("initial content\n")
        subprocess.run(["git", "add", "diff_file.txt"], cwd=temp_dir, check=True)
        subprocess.run(["git", "commit", "-m", "Initial commit"], cwd=temp_dir, check=True)

        # Modify the file that is stored in git
        with open(diff_file_path, "a") as f:
            f.write("new content\n")

        # Add a new file, do not add it to git
        not_in_git_file_path = os.path.join(temp_dir, "not_in_git.txt")
        with open(not_in_git_file_path, "w") as f:
            f.write("content outside git\n")

        # Add and commit a file but no content change
        no_change_file_path = os.path.join(temp_dir, "no_change_file.txt")
        with open(no_change_file_path, "w") as f:
            f.write("no changes here\n")
        subprocess.run(["git", "add", "no_change_file.txt"], cwd=temp_dir, check=True)
        subprocess.run(["git", "commit", "-m", "No change commit"], cwd=temp_dir, check=True)

        # Template setup for the modified file in git
        path_user_template = os.path.join(temp_dir, "templates", "diff_template.md")
        path_templates_dir = os.path.dirname(path_user_template)
        os.makedirs(path_templates_dir)
        with open(path_user_template, "w") as f:
            f.write(
                '{% set diff_output = config["current-task"]["path-modify-event"] | get_git_diff %}\n{{ diff_output }}'
            )

        # Run the renderer for the modified file
        config = get_config(path_templates_dir, diff_file_path)
        config["current-workflow"] = config["workflows"][0]
        task_diff = config["current-workflow"]["tasks"][0]
        task_diff["path-modify-event"] = diff_file_path
        config["current-task"] = task_diff
        template = os.path.basename(path_user_template)
        content_diff = render_template(task_diff, template, config, logger)

        # Check the content_diff for expected git diff
        expected_diff = (
            "--- a/diff_file.txt\n" "+++ b/diff_file.txt\n" "@@ -1 +1,2 @@\n" " initial content\n" "+new content\n"
        )
        assert expected_diff in content_diff

        config = get_config(path_templates_dir, not_in_git_file_path)
        config["current-workflow"] = config["workflows"][0]
        task_not_in_git = config["current-workflow"]["tasks"][0]
        task_not_in_git["path-modify-event"] = not_in_git_file_path
        config["current-task"] = task_not_in_git

        # Run the renderer for the not-in-git file
        content_not_in_git = render_template(task_not_in_git, template, config, logger)
        assert content_not_in_git == "\n"

        config = get_config(path_templates_dir, no_change_file_path)
        config["current-workflow"] = config["workflows"][0]
        task_no_change = config["current-workflow"]["tasks"][0]
        task_no_change["path-modify-event"] = no_change_file_path
        config["current-task"] = task_no_change

        # Run the renderer for no-change file
        content_no_change = render_template(task_no_change, template, config, logger)
        assert content_no_change == "\n"


def test_jinja_filter_file_exists():
    """
    Test of filter file_exists that can be used to render content in a jinja template
    conditional on the existence of a file.
    """
    logger = Logger(__name__, level=logging.DEBUG, log_file=f"{os.getcwd()}/debug.log")
    with tempfile.TemporaryDirectory() as temp_dir:
        existing_file_path = os.path.join(temp_dir, "existing_file.txt")
        non_existing_file_path = os.path.join(temp_dir, "non_existing_file.txt")

        # Create the existing file
        with open(existing_file_path, "w") as f:
            f.write("This file exists.")

        # Template for checking file existence
        path_user_template = os.path.join(temp_dir, "templates", "existence_template.md")
        path_templates_dir = os.path.dirname(path_user_template)
        os.makedirs(path_templates_dir)
        with open(path_user_template, "w") as f:
            f.write(
                '{% if config["current-task"]["path-modify-event"] | file_exists %}'
                "File exists: True"
                "{% else %}"
                "File exists: False"
                "{% endif %}"
            )

        # Run the renderer for the existing file
        # task_existing = get_task(path_templates_dir, existing_file_path)
        config = get_config(path_templates_dir, existing_file_path)
        config["current-workflow"] = config["workflows"][0]
        task_existing = config["current-workflow"]["tasks"][0]
        task_existing["path-modify-event"] = existing_file_path
        config["current-task"] = task_existing
        template = os.path.basename(path_user_template)
        content_existing = render_template(task_existing, template, config, logger)
        assert content_existing.strip() == "File exists: True"

        # Run the renderer for the non-existing file
        config = get_config(path_templates_dir, non_existing_file_path)
        config["current-workflow"] = config["workflows"][0]
        task_non_existing = config["current-workflow"]["tasks"][0]
        task_non_existing["path-modify-event"] = non_existing_file_path
        config["current-task"] = task_non_existing
        content_non_existing = render_template(task_non_existing, template, config, logger)
        assert content_non_existing.strip() == "File exists: False"


def test_jinja_filter_dirname_and_basename():
    """
    Test dirname and basename filters that process paths in Jinja templates.
    """
    logger = Logger(__name__, level=logging.DEBUG, log_file=f"{os.getcwd()}/debug.log")
    with tempfile.TemporaryDirectory() as temp_dir:
        # Test dirname
        path = os.path.join(temp_dir, "subfolder", "file.txt")
        dirname_result = dirname(path)
        assert dirname_result == os.path.join(temp_dir, "subfolder")

        # Test basename
        basename_result = basename(path)
        assert basename_result == "file.txt"


def test_select_task():
    """
    Test select_task function to retrieve tasks based on ID.
    """

    config = get_config2()
    config["current-workflow"] = config["workflows"][0]
    # Test for a valid task ID
    selected_task = select_task(config, "task1")
    assert selected_task["name"] == "First Task"

    # Test for an invalid task ID
    try:
        select_task(config, "task3")
        assert False, "Expected TaskNotFoundError was not raised."
    except Exception as err:
        assert isinstance(err, TaskNotFoundError)


def test_get_task_attribute():
    """
    Test get_task_attribute function to retrieve an attribute from a task.
    """

    config = get_config2()
    config["current-workflow"] = config["workflows"][0]
    config["current-task"] = config["current-workflow"]["tasks"][0]

    # Test retrieving an existing attribute
    priority = get_task_attribute(config, "task1", "priority")
    assert priority == "High"

    # Test retrieving a non-existing attribute
    try:
        get_task_attribute(config, "task1", "deadline")
        assert False, "Expected AttributeNotFoundError was not raised."
    except Exception as err:
        assert isinstance(err, AttributeNotFoundError)

    # Test with a non-existing task ID
    try:
        get_task_attribute(config, "task3", "priority")
        assert False, "Expected TaskNotFoundError was not raised."
    except Exception as err:
        assert isinstance(err, TaskNotFoundError)

    # template_env.filters["get_task_attribute"] = get_task_attribute
    # template_env.globals["raise"] = raise_helper

    # logger.debug(f"Get Jinja template: {template_path}")
    # template = template_env.get_template(template_path)
    # context = {"config": config, "task": task, "workflow": config["current-workflow"]}
    # return template.render(context)


def test_raise_in_jinja_template():
    """
    Test raising an error in a Jinja template using the raise_helper function.
    """
    logger = Logger(__name__, level=logging.DEBUG, log_file=f"{os.getcwd()}/debug.log")
    config = get_config2()
    config["current-workflow"] = config["workflows"][0]
    task = config["current-workflow"]["tasks"][0]
    config["current-task"] = task

    with tempfile.TemporaryDirectory() as temp_dir:
        config["jinja"] = {"search_path": temp_dir}
        path_template = os.path.join(temp_dir, "templates", "whatever.md")
        path_templates_dir = os.path.dirname(path_template)
        os.makedirs(path_templates_dir)
        template_string = '{{ raise("uh oh...") }}'
        with open(path_template, "w") as f:
            f.write(template_string)
        try:
            content = render_template(task, "whatever.md", config, logger)
            assert content == ""
            assert False, "Expected RuntimeError was not raised."
        except RuntimeError as err:
            assert str(err) == "uh oh..."



from coauthor.utils.jinja import render_content
import os


def read_file(config, logger):
    logger.debug(f"config: {config}")
    task = config["current-task"]
    path = task["path-modify-event"]
    logger.info(f"Reading file from {path}")
    with open(path, "r", encoding="utf-8") as file:
        content = file.read()
    task["content"] = content


def write_file(config, logger):
    logger.debug(f"config: {config}")
    task = config["current-task"]
    if "path" in task:
        path = render_content(task, task["path"], config, logger)
    else:
        path = task["path-modify-event"]

    directory = os.path.dirname(path)
    if not os.path.exists(directory):
        os.makedirs(directory)
        logger.warning(f"Directory {directory} did not exist, created it.")

    logger.info(f"Writing to file at {path}")

    content = render_content(task, task["content"], config, logger)
    with open(path, "w", encoding="utf-8") as file:
        file.write(content)



#!/usr/bin/python
#
# Example Code for Linux module `git_remote.py`
#

from __future__ import annotations
import os
from ansible.module_utils.basic import AnsibleModule


def main():
    module = AnsibleModule(
        argument_spec=dict(
            dest=dict(type="path", required=True),
            name=dict(type="str", required=True),
            url=dict(type="str", required=True),
            state=dict(
                type="str",
                required=False,
                choices=["present", "absent"],
                default="present",
            ),
        ),
        supports_check_mode=True,
    )

    dest = module.params["dest"]
    remote_name = module.params["name"]
    remote_url = module.params["url"]
    state = module.params["state"]

    result = dict(changed=False)

    git_cmd = "git -C {}".format(dest)

    if state == "present":
        # Check if remote already exists
        check_remote_command = f"{git_cmd} remote get-url {remote_name}"
        rc, stdout, stderr = module.run_command(check_remote_command)

        if rc != 0:
            # Remote does not exist, add it
            add_remote_command = f"{git_cmd} remote add {remote_name} {remote_url}"
            if module.check_mode:
                result["changed"] = True
            else:
                module.run_command(add_remote_command)
                result["changed"] = True
        else:
            # Remote exists, check url
            existing_url = stdout.strip()
            if existing_url != remote_url:
                update_remote_command = f"{git_cmd} remote set-url {remote_name} {remote_url}"
                if module.check_mode:
                    result["changed"] = True
                else:
                    module.run_command(update_remote_command)
                    result["changed"] = True

    elif state == "absent":
        # Check if remote exists
        check_remote_command = f"{git_cmd} remote get-url {remote_name}"
        rc, stdout, stderr = module.run_command(check_remote_command)

        if rc == 0:
            # Remote exists, remove it
            remove_remote_command = f"{git_cmd} remote remove {remote_name}"
            if module.check_mode:
                result["changed"] = True
            else:
                module.run_command(remove_remote_command)
                result["changed"] = True

    module.exit_json(**result)


if __name__ == "__main__":
    main()



import googleapiclient.discovery
import yaml
from youtube_transcript_api import YouTubeTranscriptApi
from youtube_transcript_api.formatters import PrettyPrintFormatter
from coauthor.utils.match_utils import regex_content_match_named_group
import os


def get_youtube_video_info(config, logger):
    """
    Retrieves and updates the task configuration with the YouTube video's
    transcript information.

    :param config: Configuration object containing task details
    :param logger: Logger for logging information
    :return: None
    """
    workflow = config["current-workflow"]
    task = config["current-task"]
    video_id = regex_content_match_named_group(workflow, "video_id", logger)
    task["video"] = {}
    task["video"]["id"] = video_id
    logger.debug(f"video_id: {video_id}")
    task["video"]["transcript"] = get_youtube_transcript(config, video_id, logger)
    task["video"]["info"] = get_youtube_metadata(config, video_id, logger)


def get_youtube_transcript(config, video_id, logger):
    """
    Retrieves the transcript of a YouTube video based on the video ID. If a
    transcript file already exists, it uses the contents of this file instead of
    calling the YouTube API.

    :param config: Configuration object containing task details
    :param video_id: The YouTube video ID
    :param logger: Optional logger for logging information
    :return: The transcript of the video as a YAML object
    """
    task = config["current-task"]
    path = task["path-modify-event"]
    path_transcript = path.rsplit(".", 1)[0] + ".transcript.yml"

    if not os.path.exists(path_transcript):
        logger.info(f'Fetching transcript for video ID: "{video_id}"')
        ytt_api = YouTubeTranscriptApi()
        languages = task.get("languages", ["en"])

        try:
            logger.debug(f"Fetching transcript for video {video_id} in languages: {languages}")
            transcript = ytt_api.fetch(video_id, languages=languages)
            formatter = PrettyPrintFormatter()

            text_formatted = formatter.format_transcript(transcript)

            # Now we can write it out to a file.
            with open(path_transcript, "w", encoding="utf-8") as yaml_file:
                yaml.dump(yaml.safe_load(text_formatted), yaml_file)

            logger.info(f"Transcript successfully retrieved for video {video_id} and written to {path_transcript}")

        except Exception as exception_error:  # pylint: disable=broad-exception-caught
            logger.error(f"Error retrieving transcript for video {video_id}: {exception_error}")
            return None

    logger.info(f"Reading video {video_id} transcript file {path_transcript}")
    with open(path_transcript, "r", encoding="utf-8") as transcript_file:
        return yaml.safe_load(transcript_file)


def get_youtube_metadata(config, video_id, logger):
    """
    Retrieves metadata for a YouTube video based on its video ID and updates the task configuration.
    If a metadata file already exists, it uses the contents of this file instead of calling the YouTube API.

    :param config: Configuration object containing task details
    :param video_id: The YouTube video ID
    :param logger: Logger for logging information
    :return: Dictionary containing video metadata
    """
    task = config["current-task"]
    path = task["path-modify-event"]
    path_info = path.rsplit(".", 1)[0] + ".info.yml"

    if not os.path.exists(path_info):
        logger.info(f'Fetching metadata for video ID: "{video_id}"')

        api_key = os.getenv("COAUTHOR_YOUTUBE_API_KEY")
        if not api_key:
            logger.error("COAUTHOR_YOUTUBE_API_KEY environment variable not set. No video info will be available.")
            return {}

        api_service_name = "youtube"
        api_version = "v3"

        youtube = googleapiclient.discovery.build(api_service_name, api_version, developerKey=api_key)

        request = youtube.videos().list(
            part="contentDetails,id,liveStreamingDetails,player,recordingDetails,snippet,statistics,status,topicDetails",
            id=video_id,
        )

        response = request.execute()

        with open(path_info, "w", encoding="utf-8") as info_file:
            yaml.dump(response["items"][0], info_file)

        logger.info(f"Metadata for video {video_id} successfully retrieved and written to {path_info}")

    logger.info(f"Reading metadata for video {video_id} from file {path_info}")
    with open(path_info, "r", encoding="utf-8") as info_file:
        return yaml.safe_load(info_file) or {}



# pylint: disable=broad-exception-caught
"""
Module for constructing AI messages based on configuration and templates.

This module provides functions to build lists of messages for AI interactions,
handling initial system and user messages, as well as additional messages
from various sources like frontmatter, files, or direct content.
"""

import os
import json
from jinja2 import Template
from coauthor.utils.jinja import render_template, template_exists, prompt_template_path, render_content
from coauthor.utils.match_utils import file_submit_to_ai
from coauthor.utils.markdown import get_frontmatter_nested_value
from coauthor.modules.tools import execute_tool
from coauthor.utils.config import get_projects


def ai_messages(config, logger):
    """
    Construct and return a list of AI messages based on the configuration.

    This function initializes a list of messages, adds initial system and user
    messages, and then appends any additional messages specified in the task.

    Args:
        config (dict): Configuration dictionary containing the current task.
        logger (Logger): Logger instance for logging messages.

    Returns:
        list: A list of message dictionaries, each with 'role' and 'content'.
    """
    messages = []
    if not add_initial_messages(messages, config, logger):
        return []
    add_additional_messages(messages, config, logger)
    return messages


def add_initial_messages(messages, config, logger):
    """
    Add initial system and user messages to the messages list.

    This function retrieves and renders templates for system and user roles,
    appending them to the messages list if content is available.

    Args:
        messages (list): List to which messages will be appended.
        config (dict): Configuration dictionary containing the current task.
        logger (Logger): Logger instance for logging messages.

    Returns:
        bool: True if both messages were added successfully, False otherwise.
    """
    task = config["current-task"]
    for role in ["system", "user"]:
        path_template = prompt_template_path(config, f"{role}.md", logger)
        task[f"{role}_template_path"] = path_template
        content = ai_message_content(config, logger, path_template, role)
        if content:
            messages.append({"role": role, "content": content})
        else:
            logger.error("Message content missing!")
            return False
    return True


def add_additional_messages(messages, config, logger):
    """
    Add additional user messages to the messages list based on task configuration.

    This function processes additional messages specified in the task, rendering
    their content and appending them as user messages.

    Args:
        messages (list): List to which messages will be appended.
        config (dict): Configuration dictionary containing the current task.
        logger (Logger): Logger instance for logging messages.
    """
    task = config["current-task"]
    if "messages" not in task:
        return
    for msg in task["messages"]:
        logger.debug(f"message: {msg}")
        items = get_message_items(msg, task, config, logger)
        for item in items:
            if item is not None:
                if "frontmatter" in msg:
                    task["frontmatter-item"] = item
                elif "files" in msg:
                    task["user-message-context-file"] = item
            content = get_additional_message_content(msg, config, logger, task)
            if content:
                logger.info(f"Adding user message: {msg}")
                messages.append({"role": "user", "content": content})
            else:
                logger.error("Missing content for additional user message")


def get_message_items(msg, task, config, logger):
    """
    Retrieve items based on the message configuration.

    If the message specifies a frontmatter key, this function fetches the
    corresponding list from the task's frontmatter. If 'files' is specified,
    it renders the directory path and lists all files in that directory.

    Args:
        msg (dict): Message dictionary potentially containing 'frontmatter' or 'files' key.
        task (dict): Task dictionary containing frontmatter data.
        config (dict): Configuration dictionary.
        logger (Logger): Logger instance for logging messages.

    Returns:
        list: List of items from frontmatter or files, or [None] if not specified.
    """
    if "frontmatter" in msg:
        frontmatter_key = msg["frontmatter"]
        frontmatter_list = get_frontmatter_nested_value(task["path-modify-event"], frontmatter_key)
        if frontmatter_list is None or not frontmatter_list:
            return []
        return frontmatter_list
    if "files" in msg:
        try:
            template_path = msg["files"]
            dir_path = render_content(task, template_path, config, logger)
            logger.debug(f"Looking for files in dir_path: {dir_path}")
        except Exception as exception_error:
            logger.error(f"Failed to render files path: {exception_error}")
            return []
        if not os.path.isdir(dir_path):
            logger.error(f"Directory does not exist: {dir_path}")
            return []
        files = []
        for root, _dirs, filenames in os.walk(dir_path):
            for filename in filenames:
                full = os.path.join(root, filename)
                files.append(full)
        logger.debug(f"Files to include as user message: {', '.join(files)}")
        return files
    return [None]


def get_additional_message_content(msg, config, logger, task):
    """
    Generate content for an additional message.

    This function retrieves content either from a file template or by rendering
    a Jinja2 template string provided in the message.

    Args:
        msg (dict): Message dictionary with 'file' or 'content' for sourcing.
        config (dict): Configuration dictionary.
        logger (Logger): Logger instance for logging messages.
        task (dict): Current task dictionary.

    Returns:
        str or None: The rendered message content, or None if rendering fails.
    """
    content = None
    if "file" in msg:
        path_template = msg["file"]
        content = ai_message_content(config, logger, path_template, "user")
    elif "content" in msg:
        template_string = msg["content"]
        try:
            template = Template(template_string)
            content = template.render(task=task, config=config)
        except Exception as exception_error:
            logger.error(f"Failed to render content: {exception_error}")
            content = None
    else:
        logger.error("Message has neither 'file' nor 'content'")
    return content


def ai_message_content(config, logger, path_template, system_or_user):
    """
    Retrieve or render content for an AI message.

    This function checks for template existence and renders it, falls back to
    task data, or uses file content for user messages in modify events.

    Args:
        config (dict): Configuration dictionary containing the current task.
        logger (Logger): Logger instance for logging messages.
        path_template (str): Path to the template file.
        system_or_user (str): Role of the message ('system' or 'user').

    Returns:
        str or None: The message content, or None if not found.
    """
    task = config["current-task"]
    message_content = None
    if template_exists(task, path_template, config, logger):
        message_content = render_template(task, path_template, config, logger)
    elif system_or_user in task:
        message_content = task[system_or_user]
    elif "path-modify-event" in task and system_or_user == "user":
        logger.info(f"Using the file {task['path-modify-event']} as the user message")
        message_content = file_submit_to_ai(config, logger)
    return message_content


def insert_projects_status_message(messages, config, logger):
    """
    Insert a user message with project information including Git status into the messages list.

    This function retrieves project details, fetches modified files for each project using
    the list_modified_files tool, constructs a JSON-formatted string of project infos,
    and inserts it as a user message before the first existing user message.

    Args:
        messages (list): List of message dictionaries to modify.
        config (dict): Configuration dictionary.
        logger (Logger): Logger instance for logging messages.
    """
    projects = config.get("all_projects", get_projects(config))
    project_infos = []
    for project in projects:
        project_name = project.get("name")
        project_path = os.path.expanduser(project.get("path", os.getcwd()))
        info = {
            "name": project_name,
            "type": project.get("type"),
            "description": project.get("description"),
        }
        if project_path == "none":
            info["note"] = "This project has no files."
        else:
            try:
                modified_files = execute_tool(config, "list_modified_files", {"project_name": project_name}, logger)
            except Exception as exception_error:
                logger.error(f"Failed to get modified files for {project_name}: {exception_error}")
                modified_files = []
            info["modified_files"] = modified_files
        project_infos.append(info)
    projects_content = f"Information about the projects:\n{json.dumps(project_infos, indent=2)}"
    projects_message = {"role": "user", "content": projects_content}
    insert_index = 0
    for i, msg in enumerate(messages):
        if msg["role"] == "user":
            insert_index = i
            break
    else:
        insert_index = len(messages)
    messages.insert(insert_index, projects_message)



import os
import logging
from coauthor.utils.logger import Logger
from coauthor.modules.workflow import initialize_workflows
from coauthor.utils.config import get_config
import tempfile
import pytest
import threading
import time
import argparse


class MockLogger:
    def info(self, message):
        print(message)

    def debug(self, message):
        print(message)

    def warning(self, message):
        print(message)


def test_initialize_workflows_no_workflows():
    """
    Test the initialize_workflows function when there are no workflows in the config.
    A warning should be logged and no further action taken.
    """
    # Create a mock logger
    mock_logger = MockLogger()

    config = get_config()
    if "workflows" in config:
        del config["workflows"]

    initialize_workflows(config=config, logger=mock_logger)

    # Check that the warning message was logged
    warning_logged = False
    expected_message = "No workflows found in the configuration. The program will now exit as there is nothing to do."

    # Simulate the functionality of checking logger output
    def mock_warning(message):
        nonlocal warning_logged
        if message == expected_message:
            warning_logged = True

    mock_logger.warning = mock_warning

    # Call the function
    initialize_workflows(config=config, logger=mock_logger)

    # Assert the message was logged
    assert warning_logged


def test_initialize_workflows_no_scan_directories():
    """
    Test the initialize_workflows function when there are no scan directories configured.
    Expect debug information about lack of scan directories.
    """
    # Create a mock logger
    mock_logger = MockLogger()

    path = os.path.join(os.path.dirname(__file__), "data", "coauthor_translation2.yml")
    config = get_config(path=path)

    # Ensure the scan paths are empty
    if "workflows" in config:
        config["workflows"][0]["scan"] = {"filesystem": {"paths": []}}

    initialize_workflows(config=config, logger=mock_logger, trigger_scan=True)


def test_read_file_write_file():
    """
    Test the workflow that reads a file and then writes to it using specific content patterns.

    The workflow is configured to scan files with `.md` extension, looking for content that matches
    the pattern ".*ping.*". Upon finding such files, the "read_file" task reads the content,
    and the "write_file" task modifies the content to append "=pong".
    """
    path = os.path.join(os.path.dirname(__file__), "data", "coauthor_translation2.yml")
    with tempfile.TemporaryDirectory() as temp_dir:
        # Create a temporary directory with a subdirectory
        sub_dir = os.path.join(temp_dir, "subdir")
        os.makedirs(sub_dir, exist_ok=True)

        # Define the file path and write "ping" into it
        file_path = os.path.join(sub_dir, "testfile.md")
        with open(file_path, "w") as file:
            file.write("ping")

        # Define the configuration for the workflow
        config = get_config(path=path)
        workflow = config["workflows"][0]
        workflow["scan"] = {"filesystem": {"paths": [temp_dir]}}

        # Initialize a Logger instance for debugging
        logger = Logger(__name__, level=logging.DEBUG, log_file=f"{os.getcwd()}/debug.log")

        # Call the `initialize_workflows` to process the file
        initialize_workflows(config=config, logger=logger, trigger_scan=True)

        # Check if the content of the file was updated correctly
        with open(file_path, "r") as file:
            content = file.read()

        # Assert if the content contains "ping=pong"
        assert "ping=pong" in content


# test that we can watch and stop watching
def test_initialize_workflow_watch_and_stop():
    def create_parser():
        parser = argparse.ArgumentParser(description="Process arguments to select steps")
        parser.add_argument(
            "--watch",
            action="store_true",
            help="Flag to enable watch mode",
        )
        parser.add_argument(
            "--scan",
            action="store_true",
            help="Flag to enable scan mode",
        )
        return parser

    def write_stop_file(path_stop_file, logger):
        logger.info(f"sleep 1 seconds")
        time.sleep(1)

        logger.info(f"Write stop file")
        with open(path_stop_file, "w") as f:
            f.write("stop")

    # Create a mock logger
    mock_logger = MockLogger()

    # Check that the warning message was logged
    warning_logged = False
    info_logged = True
    expected_message_1 = "Stop file found!"
    expected_message = (
        "No directories to watch specified in workflows. The program will now exit as there is nothing to do."
    )

    def mock_warning(message):
        nonlocal warning_logged
        if message == expected_message:
            warning_logged = True

    def mock_info(message):
        nonlocal info_logged
        if message == expected_message_1:
            info_logged = True

    mock_logger.warning = mock_warning
    mock_logger.info = mock_info

    with tempfile.TemporaryDirectory() as temp_dir:
        logger = Logger("test_initialize_workflow_watch", level=logging.DEBUG, log_file=f"{os.getcwd()}/debug.log")
        path = os.path.join(os.path.dirname(__file__), "data", "coauthor_translation2.yml")
        config = get_config(path=path)
        config["current-workflow"] = config["workflows"][0]
        config["current-workflow"]["watch"] = {"filesystem": {"paths": [temp_dir]}}
        config["current-task"] = config["current-workflow"]["tasks"][0]
        parser = create_parser()
        args = parser.parse_args(["--watch"])
        config["args"] = args
        path_stop_file = os.path.join(temp_dir, "stop")

        x = threading.Thread(target=write_stop_file, args=(path_stop_file, logger), daemon=True)
        x.start()
        initialize_workflows(config, mock_logger)
        x.join()
        assert info_logged

        config["current-workflow"]["watch"] = {"filesystem": {"paths": []}}
        x = threading.Thread(target=write_stop_file, args=(path_stop_file, logger), daemon=True)
        x.start()
        initialize_workflows(config, mock_logger)
        x.join()

        assert warning_logged



"""
This module serves as the entry point for processing command-line arguments
related to configuration and workflow initialization. It configures logging,
retrieves configuration details, and initializes workflows based on the parsed
command-line arguments.
"""

import argparse
import logging
import os
import sys
import yaml
from coauthor.utils.logger import Logger
from coauthor.utils.config import get_config, save_config_dump
from coauthor.modules.workflow import initialize_workflows
from coauthor.utils.export import export
from coauthor import __version__


def main():
    """
    Parses command-line arguments, sets up logging, retrieves configuration,
    and initializes workflows.

    The function uses argparse to handle command-line arguments, sets up a logger based
    on the provided or default log level and log file path, and retrieves configuration
    details. It then initializes workflows using the collected configuration and logger.

    Command-line arguments:

    --config_path: Path to the configuration file.

    --profile: Name of profile (e.g., obsidian, python, hugo).

    --watch: Flag to enable watch mode.
    --scan: Flag to enable scan mode.
    --debug: Flag to set log level to DEBUG.
    --log_file: Path to the log file.
    --export: Flag to enable export mode.
    """
    parser = argparse.ArgumentParser(description="Process arguments to select steps")
    parser.add_argument(
        "--config_path",
        type=str,
        help="Path to the configuration file",
    )
    version_str = f"coauthor {__version__}"
    parser.add_argument("--version", action="version", version=version_str)
    parser.add_argument(
        "--profile",
        type=str,
        help="Name of profile e.g. obsidian, python, hugo, ansible-collection",
    )
    parser.add_argument(
        "--watch",
        action="store_true",
        help="Flag to enable watch mode",
    )
    parser.add_argument(
        "--scan",
        action="store_true",
        help="Flag to enable scan mode",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Flag to set log level to DEBUG",
    )
    parser.add_argument(
        "--log_file",
        type=str,
        help="Path to the log file",
    )
    parser.add_argument(
        "--notify",
        action="store_true",
        help="Flag to enable desktop notifications",
    )
    parser.add_argument(
        "--export",
        action="store_true",
        help="Flag to enable export mode",
    )
    args = parser.parse_args()

    log_level = logging.DEBUG if args.debug else logging.INFO
    if args.log_file:
        log_file = args.log_file
    else:
        log_file = f"{os.getcwd()}/coauthor.log"
    logger = Logger(__name__, log_file=log_file, level=log_level)
    config = get_config(logger=logger, args=args)
    if args.debug:
        save_config_dump(config, logger)
    config["args"] = args
    if args.export:
        export(config=config, logger=logger)
    else:
        initialize_workflows(config, logger)


if __name__ == "__main__":  # pragma: no cover
    main()



import os
import datetime
import yaml
import json


def get_log_dir():
    log_dir = os.getenv("COAUTHOR_AI_LOG_DIR")
    if not log_dir:
        log_dir = os.path.join(os.getcwd(), "tmp")
        os.makedirs(log_dir, exist_ok=True)
    return log_dir


def get_ai_request_id():
    counter_file_path = os.path.join(get_log_dir(), ".ai-prompt-counter")
    if not os.path.exists(counter_file_path):
        return 0
    with open(counter_file_path, "r", encoding="utf-8") as counter_file:
        return int(counter_file.read())


def next_ai_request_id():
    current_id = get_ai_request_id()
    new_id = current_id + 1
    counter_file_path = os.path.join(get_log_dir(), ".ai-prompt-counter")
    with open(counter_file_path, "w", encoding="utf-8") as counter_file:
        counter_file.write(str(new_id))
    return get_ai_request_id()


def get_ai_workflow_id():
    counter_file_path = os.path.join(get_log_dir(), ".ai-workflow-counter")
    if not os.path.exists(counter_file_path):
        return 0
    with open(counter_file_path, "r", encoding="utf-8") as counter_file:
        return int(counter_file.read())


def next_ai_workflow_id():
    current_id = get_ai_workflow_id()
    new_id = current_id + 1
    counter_file_path = os.path.join(get_log_dir(), ".ai-workflow-counter")
    with open(counter_file_path, "w", encoding="utf-8") as counter_file:
        counter_file.write(str(new_id))
    return get_ai_workflow_id()


def get_or_create_request_dir(workflow_id):
    log_dir = get_log_dir()
    workflow_str = f"{workflow_id:08d}"
    request_id = get_ai_request_id()
    request_str = f"{request_id:08d}"
    dir_path = os.path.join(log_dir, workflow_str, request_str)
    os.makedirs(dir_path, exist_ok=True)
    return dir_path, request_id


def save_messages(messages, model, kwargs, logger, workflow_id):
    dir_path, _ = get_or_create_request_dir(workflow_id)
    if not dir_path:
        return

    idx = 1  # counter for txt files
    processed_messages = []
    for msg in messages:
        processed_msg = msg.copy()
        content = processed_msg["content"]
        if isinstance(content, str):
            try:
                processed_msg["content"] = json.loads(content)
            except json.JSONDecodeError:
                # It's not JSON, save to file
                txt_path = os.path.join(dir_path, f"{idx}.txt")
                with open(txt_path, "w", encoding="utf-8") as f:
                    f.write(content)
                logger.debug(f"Message content written to: {txt_path}")

                # Truncate
                lines = content.splitlines()
                if len(lines) > 3:
                    truncated = "\n".join(lines[:3]) + "\n(lines deleted)"
                else:
                    truncated = "\n".join(lines)
                processed_msg["content"] = truncated
                idx += 1
        processed_messages.append(processed_msg)

    request = {"messages": processed_messages, "model": model, "kwargs": kwargs}  # , "kwargs": kwargs
    request_path = os.path.join(dir_path, "request.yml")
    with open(request_path, "w", encoding="utf-8") as file:
        yaml.dump(request, file, default_flow_style=False, allow_unicode=True)
    logger.debug(f"Messages → {request_path}")


def save_response(response, logger, workflow_id):  # , duration=None
    # def save_response(config, messages, model, response, logger, duration=None):
    # response = [[tc.model_dump() for tc in tool_calls] if tool_calls else None, content]
    dir_path, _message_id = get_or_create_request_dir(workflow_id)
    response_file_path = os.path.join(dir_path, "response.yml")
    with open(response_file_path, "w", encoding="utf-8") as file:
        yaml.dump(response.model_dump(), file, default_flow_style=False, allow_unicode=True)
    logger.info(f"Response → {response_file_path}")
    # data = {
    #     "messages": [{"role": msg["role"], "tokens": len(msg["content"].split())} for msg in messages],  # Count words
    #     "model": model,
    #     "response": response,
    #     "task": task,
    #     "date": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
    #     "duration": duration,  # Add the duration to the data
    # }

    # task["ai-log-file"] = response_file_path

    # json_mode = task.get("json", False)
    # if json_mode:
    #     file_path = os.path.join(dir_path, f"{_message_id}-response.json")
    #     with open(file_path, "w", encoding="utf-8") as f:
    #         f.write(response)
    # else:
    #     file_path = os.path.join(dir_path, f"{_message_id}-response.txt")
    #     with open(file_path, "w", encoding="utf-8") as f:
    #         f.write(response)
    # logger.info(f"Response written to {'JSON' if json_mode else 'TXT'} file: {file_path}")



import yaml
import os
import importlib.resources
import copy


def deep_merge(d1, d2):
    for key, value in d2.items():
        if key in d1 and isinstance(d1[key], dict) and isinstance(value, dict):
            deep_merge(d1[key], value)
        else:
            d1[key] = value
    return d1


def read_config(file_path, logger=None):
    if logger:
        logger.info(f"Reading configuration from {file_path}")
    with open(file_path, "r", encoding="utf-8") as file:
        return yaml.safe_load(file)


def get_config_path(config_filename=".coauthor.yml", search_dir=os.getcwd()):
    traversed_paths = []
    while True:
        potential_path = os.path.join(search_dir, config_filename)
        if os.path.exists(potential_path):
            return potential_path, traversed_paths
        traversed_paths.append(search_dir)
        parent_dir = os.path.dirname(search_dir)
        if parent_dir == search_dir:
            break
        search_dir = parent_dir

    home_dir = os.path.expanduser("~")
    home_path = os.path.join(home_dir, config_filename)
    if os.path.exists(home_path):
        traversed_paths.append(home_dir)
        return home_path, traversed_paths

    return None, traversed_paths


def get_default_config():
    config = {
        "jinja": {"search_path": ".coauthor/templates"},
        "agent": {
            "api_key_var": "OPENAI_API_KEY",
            "api_url_var": "OPENAI_API_URL",
            "model": "x-ai/grok-4",
        },
        "file-watcher": {"ignore-folders": ["__pycache__", ".obsidian", ".git"]},
    }
    return config


def get_projects(config):
    projects = config.get("projects", [])
    config_name = config.get("name")
    if config_name and not any(p.get("name") == config_name for p in projects):
        projects += [config]
    return projects


def expand_paths(conf):
    if "path" in conf:
        conf["path"] = os.path.expanduser(conf["path"])
    if "projects" in conf:
        for proj in conf["projects"]:
            expand_paths(proj)


def save_config_dump(config, logger=None):
    dump_path = os.path.join(os.getcwd(), ".coauthor_dump.yml")
    with open(dump_path, "w", encoding="utf-8") as dump_file:
        yaml.safe_dump(config, dump_file, default_flow_style=False)
        if logger:
            logger.info(f"Dumped configuration to {dump_path}")


def get_config(path=None, logger=None, config_filename=".coauthor.yml", search_dir=os.getcwd(), args=None):
    config = {}
    config_path = None
    if args and hasattr(args, "config_path") and args.config_path:
        config_path = args.config_path
    elif args and hasattr(args, "profile") and args.profile:
        profile = args.profile
        profile_path = importlib.resources.files("coauthor.profiles").joinpath(profile).joinpath("config.yml")
        config = get_default_config()
        deep_merge(config, read_config(profile_path, logger))
        expand_paths(config)
        return config
    if not config_path:
        if path:
            config_path = path
        else:
            config_path, searched_paths = get_config_path(config_filename, search_dir)
            if not config_path:
                if logger:
                    logger.warning(f"Configuration file not found. Searched directories: {', '.join(searched_paths)}")
                config_path = os.path.join(os.getcwd(), config_filename)
                config = get_default_config()
                with open(config_path, "w", encoding="utf-8") as file:
                    if logger:
                        logger.debug(f"Dump config to YAML file {config_path}")
                    yaml.safe_dump(config, file)

                if logger:
                    logger.info(f"Created default configuration file at {config_path}")
    config = get_default_config()
    deep_merge(config, read_config(config_path, logger))
    expand_paths(config)
    if "projects" in config:
        for proj in config["projects"]:
            if "profile" in proj:
                logger.info(f"Applying profile {proj['profile']} config to project {proj['name']}")
                profile = proj["profile"]
                profile_path = importlib.resources.files("coauthor.profiles").joinpath(profile).joinpath("config.yml")
                profile_config = read_config(profile_path, logger)
                # Flatten if single workflow
                workflows = profile_config.get("workflows", [])
                if len(workflows) == 1:
                    workflow = workflows[0]
                    deep_merge(profile_config, workflow)
                # Apply profile_args if present
                if "profile_args" in proj:
                    deep_merge(profile_config, proj["profile_args"])
                    del proj["profile_args"]
                local = copy.deepcopy(proj)
                proj.clear()
                default = get_default_config()
                deep_merge(proj, default)
                deep_merge(proj, profile_config)
                deep_merge(proj, local)
        expand_paths(config)
        all_projects = get_projects(config)
        for proj in config["projects"]:
            proj["all_projects"] = all_projects
    return config


def get_jinja_config(config):
    if "jinja" in config:
        return config["jinja"]
    config_jinja = {"search_path": ".coauthor/templates"}
    return config_jinja



import os
import re


def file_submit_to_ai(config, logger):
    """
    Return the content of the file if it requires AI processing based on path and content matching criteria.

    Parameters:
    - config (dict): Configuration dictionary containing 'current-task' and 'current-workflow'.
    - logger (Logger): Logger to log debug and warning messages.

    Returns:
    - str: The content of the file if both path and content match the specified regex patterns in
      the task.
    - None: If the content doesn't meet the regex requirements.

    The function checks if the file path and content match the given regex patterns in the task.
    """
    path = config["current-task"]["path-modify-event"]
    path_match = file_path_match(config, logger)
    content_match = file_content_match(config, logger)
    content = None
    if path_match:
        if content_match:
            with open(path, "r", encoding="utf-8") as file1:
                content = file1.read()
    logger.debug(f"file_submit_to_ai: path_match: {path_match}, content_match: {content_match}")
    return content


def regex_content_match(content, workflow, patterns_key, matches_key, logger):
    """
    Determine if the given content matches any provided regex patterns.

    Parameters:
    - content (str): The content to be validated against the regex.
    - workflow (dict): Should contain a list under 'content_patterns' with regex patterns to match
      against the content. Also updates 'content_matches' with the match object of each match.
    - patterns_key: content_patterns or path_patterns
    - matches_key: content_matches or path_matches

    Returns:
    - bool: True if the content matches any of the regex patterns, otherwise False.

    The function checks the content against each regex pattern for a match, updating the
    'content_matches' list in the workflow with the match object.
    """
    logger.debug(f"workflow: {workflow}")
    logger.debug(f"patterns_key: {patterns_key}, matches_key: {matches_key}")
    regex = workflow[patterns_key]
    workflow[matches_key] = []
    for pattern in regex:
        if isinstance(pattern, list):
            sub_patterns_all_match = True
            for sub_pattern in pattern:
                match_object = re.search(sub_pattern, content, re.IGNORECASE | re.DOTALL)
                workflow[matches_key].append(match_object)
                match = bool(match_object)
                if not match:
                    logger.debug(f"no match for sub_pattern: {sub_pattern} → sub_patterns_all_match = False")
                    sub_patterns_all_match = False
            if sub_patterns_all_match:
                logger.debug(f"regex_content_match: match True for pattern: {pattern}, content: {content}")
                return True
        else:
            match_object = re.search(pattern, content, re.IGNORECASE | re.DOTALL)
            workflow[matches_key].append(match_object)
            if match_object:
                logger.debug(f"regex_content_match: match: True for pattern: {pattern}, content: {content}")
                return True
    return False


def regex_content_match_named_group(workflow, named_group, logger):
    """
    Iterate through all "content_matches" of "workflow" and return the value of a named group if it exists in any match.

    Parameters:
    - workflow (dict): Contains 'content_matches' list with regex match results.
    - named_group (str): The named group to check within the content matches.
    - logger (Logger): Logger to log debug messages.

    Returns:
    - str: Value of the named group if found, otherwise None.
    """
    for match in workflow["content_matches"]:
        if isinstance(match, re.Match):
            if named_group in match.groupdict():
                logger.debug(f"regex_content_match_named_group: Found named group '{named_group}'")
                return match.group(named_group)
    logger.debug(f"regex_content_match_named_group: Named group '{named_group}' not found")
    return None


def file_path_match(config, logger):
    """
    Check if the file path matches the regex pattern specified in the workflow.

    Parameters:
    - config (dict): Configuration dictionary containing 'current-task' and 'current-workflow'.
    - logger (Logger): Logger to log debug and warning messages.

    Returns:
    - bool: True if the path matches the regex or if no 'path_patterns' is specified, otherwise False.
    """
    workflow = config["current-workflow"]
    path = config["current-task"]["path-modify-event"]
    if "path_patterns" in workflow:
        return regex_content_match(path, workflow, "path_patterns", "path_matches", logger)
    return True


def file_content_match(config, logger):
    """
    Check if the file content matches the regex pattern specified in the workflow configuration.

    Parameters:
    - config (dict): Configuration dictionary containing 'current-task' and 'current-workflow'.
    - logger (Logger): Logger to log debug and warning messages.

    Returns:
    - bool: True if the content matches the regex, otherwise False.
    """
    workflow = config["current-workflow"]
    path = config["current-task"]["path-modify-event"]
    if not os.path.exists(path):
        logger.warning(f"file_content_match: path {path} does not exist!")
        return False

    if "content_patterns" in workflow:
        logger.debug(f"file_content_match: content_patterns: {workflow['content_patterns']}")
        with open(path, "r", encoding="utf-8") as file:
            content = file.read()
        if regex_content_match(content, workflow, "content_patterns", "content_matches", logger):
            return True
    else:
        logger.warning(f"file_content_match: workflow has no content_patterns! So content match is false! ")
        logger.debug(f"workflow: {workflow}")
    return False


def path_new_replace(path, search, replace):
    """
    Replace a portion of the path with a new string.

    Parameters:
    - path (str): The original path string.
    - search (str): The substring to search for in the path.
    - replace (str): The string to replace the search substring with.

    Returns:
    - str: The modified path with the substitutions made.
    """
    return path.replace(search, replace)



import logging
import os
import argparse

import yaml
import time
import tempfile
from coauthor.utils.config import get_config, get_config_path, read_config, get_jinja_config
from coauthor.utils.logger import Logger


def test_get_config_path_returns_correct_coauthor_file():
    """
    Test that the `get_config_path` function correctly returns the file path
    of the '.coauthor.yml' file located in the current project root directory.
    """
    file_path = get_config_path()
    file_path_expected = os.path.join(os.getcwd(), ".coauthor.yml")
    assert file_path[0] == file_path_expected


def test_get_config_path_returns_none():
    config_path = get_config_path("whatever.yml")
    assert config_path[0] is None


def test_get_config_path_returns_file_in_home_directory():
    home_directory = os.path.expanduser("~")
    file_name = "whatever.yml"
    file_path = os.path.join(home_directory, file_name)
    with open(file_path, "w") as file:
        file.write("This is a test file created using Python.")
    config_path = get_config_path("whatever.yml", "/tmp")
    os.remove(file_path)
    assert config_path[0] == file_path


def test_get_config():
    config = get_config()  # this return the config of this project
    assert config["workflows"][0]["name"] == "coauthor"
    assert config["workflows"][0]["tasks"][0]["id"] == "python"


def test_get_config_with_logger():
    logger = Logger(__name__, level=logging.DEBUG, log_file="debug.log")
    config = get_config(None, logger)  # this return the config of this project
    assert config["workflows"][0]["name"] == "coauthor"
    assert config["workflows"][0]["tasks"][0]["id"] == "python"


def test_get_config_whatever_profile():
    path = os.path.join(os.path.dirname(__file__), "data", "coauthor.yml")
    config = get_config(path)
    assert config["profile"] == "whatever"


def test_get_config_not_found_with_logger():
    logger = Logger(__name__, level=logging.INFO, log_file="coauthor.log")
    tmp_config_file = ".coauthor_test_get_config_not_found_with_logger.yml"
    if os.path.exists(tmp_config_file):
        os.remove(tmp_config_file)
    config = get_config(logger=logger, config_filename=tmp_config_file, search_dir="/tmp")
    assert config is not None
    os.remove(tmp_config_file)


def test_config_agent_model():
    path = os.path.join(os.path.dirname(__file__), "data", "coauthor-markdown2.yml")
    config = get_config(path)
    assert config["agent"]["model"] == "llama2:latest"


def test_config_watch_directory():
    path = os.path.join(os.path.dirname(__file__), "data", "coauthor-markdown2.yml")
    config = get_config(path)
    assert config["watch_directory"] == "/whatever"


def test_config_callback():
    path = os.path.join(os.path.dirname(__file__), "data", "coauthor-markdown2.yml")
    config = get_config(path)
    assert config["callback"] == "whatever_callback"


def test_config_api_url_var():
    path = os.path.join(os.path.dirname(__file__), "data", "coauthor-markdown2.yml")
    config = get_config(path)
    assert config["agent"]["api_url_var"] == "WHATEVER_URL"


def test_config_api_key_var():
    path = os.path.join(os.path.dirname(__file__), "data", "coauthor-markdown2.yml")
    config = get_config(path)
    assert config["agent"]["api_key_var"] == "WHATEVER_KEY"


def test_config_defaults():
    cwd = os.getcwd()
    tmp_config_file = ".coauthor_test_config_defaults.yml"
    config_path = os.path.join(cwd, tmp_config_file)
    if os.path.exists(tmp_config_file):
        os.remove(tmp_config_file)
    with open(config_path, "w") as file:
        yaml.safe_dump({}, file)
    config = get_config(config_filename=tmp_config_file)

    assert config["agent"]["api_url_var"] == "OPENAI_API_URL"
    assert config["agent"]["api_key_var"] == "OPENAI_API_KEY"
    # assert config["agent"]["model"] == "openai/gpt-4o"
    assert config["agent"]["model"] == "gpt-4o-mini"
    assert config["jinja"]["search_path"] == ".coauthor/templates"
    assert config["file-watcher"]["ignore-folders"] == ["__pycache__", ".obsidian", ".git"]
    os.remove(tmp_config_file)


def test_get_config_creates_default_file_if_not_exists():

    def create_parser():
        parser = argparse.ArgumentParser(description="Process arguments to select steps")
        parser.add_argument(
            "--config_path",
            type=str,
            help="Path to the configuration file",
        )
        return parser

    cwd = os.getcwd()
    logger = Logger(__name__, level=logging.DEBUG, log_file=f"{cwd}/debug.log")
    tmp_config_file = ".coauthor_test_get_config_creates_default_file_if_not_exists.yml"
    if os.path.exists(tmp_config_file):
        os.remove(tmp_config_file)
    config_path = os.path.join(cwd, tmp_config_file)
    # config = get_config(logger=logger, config_filename=tmp_config_file)
    config = get_config(config_filename=tmp_config_file)
    assert os.path.exists(config_path)

    assert config["agent"]["api_key_var"] == "OPENAI_API_KEY"
    assert config["agent"]["api_url_var"] == "OPENAI_API_URL"
    # assert config["agent"]["model"] == "openai/gpt-4o"
    assert config["agent"]["model"] == "gpt-4o-mini"

    parser = create_parser()
    args = parser.parse_args(["--config_path", config_path])
    config = get_config(args=args)
    assert os.path.exists(config_path)
    os.remove(tmp_config_file)


def test_get_jinja_config_defaults():
    config = {}
    config_jinja = get_jinja_config(config)
    assert config_jinja == {"search_path": ".coauthor/templates"}



import tempfile
import logging
import pytest
import yaml
from unittest import mock
from coauthor.utils.logger import Logger
import os
from coauthor.modules.youtube import get_youtube_transcript, get_youtube_video_info
from youtube_transcript_api import YouTubeTranscriptApi
from youtube_transcript_api.formatters import TextFormatter
from youtube_transcript_api._errors import TranscriptsDisabled, NoTranscriptFound, VideoUnavailable
from coauthor.utils.match_utils import regex_content_match, regex_content_match_named_group


def test_youtube_transcript_retrieval():
    """
    Pytest test for get_youtube_transcript function.
    Tests retrieval of a transcript for provided YouTube video ID.
    """
    video_id = "ggyMIg8uXsw"
    logger = Logger(__name__, level=logging.DEBUG, log_file=f"{os.getcwd()}/debug.log")

    with tempfile.TemporaryDirectory() as temp_dir:
        config = {"current-task": {"path-modify-event": f"{temp_dir}/video.transcript"}}

        result_transcript = get_youtube_transcript(config, video_id, logger)
        logger.debug(f"result_transcript: {result_transcript}")

        expected_transcript_path = config["current-task"]["path-modify-event"] + ".txt"

        assert result_transcript is not None, "Expected non-None transcript."
        assert "President Trump's 90-day" in result_transcript, "Expected text not found in the transcript."

        assert os.path.exists(expected_transcript_path), "Transcript file does not exist."
        assert os.path.getsize(expected_transcript_path) > 0, "Transcript file is empty."


def test_get_youtube_video_info():
    content = "whatever\nwhatever\nyoutube_id: ggyMIg8uXsw\n\nwhatever\n@whatever: something\n\n"
    logger = Logger(__name__, level=logging.DEBUG, log_file=f"{os.getcwd()}/debug.log")

    with tempfile.TemporaryDirectory() as temp_dir:

        workflow = {}
        workflow["content_patterns"] = [[".*@whatever:.*", ".*youtube_id:\s*(?P<video_id>\S+).*"]]
        config = {"current-task": {"path-modify-event": f"{temp_dir}/video.transcript"}}
        config["current-workflow"] = workflow
        result = regex_content_match(content, workflow, "content_patterns", "content_matches", logger)

        assert result is True, "Expected result to be True."
        assert "content_matches" in workflow, "Expected 'content_matches' key in workflow."
        logger.debug(f"content_matches: {workflow['content_matches']}")
        assert len(workflow["content_matches"]) == 2, "Expected two items in 'content_matches'."

        for match in workflow["content_matches"]:
            logger.debug(f"match: {match}")

        get_youtube_video_info(config, logger)

        assert (
            "video" in config["current-task"] and config["current-task"]["video"]["id"] == "ggyMIg8uXsw"
        ), "Expected video id to be 'ggyMIg8uXsw'."
        assert "transcript" in config["current-task"]["video"], "Expected transcript"

        transcript_yml = config["current-task"]["video"]["transcript"]
        transcript_txt = yaml.dump(transcript_yml, default_flow_style=False)
        logger.debug(f"transcript_txt: {transcript_txt}")
        assert "President Trump's 90-day" in transcript_txt, "Expected 90-day"

        get_youtube_video_info(config, logger)

        logger.debug(f"info: {config['current-task']['video']['info']}")
        assert "info" in config["current-task"]["video"], "Expected info"
        assert "snippet" in config["current-task"]["video"]["info"], "Expected snippet"
        assert "title" in config["current-task"]["video"]["info"]["snippet"], "Expected snippet"
        assert (
            "The EU just FOUND OUT." in config["current-task"]["video"]["info"]["snippet"]["title"]
        ), "Expected The EU just FOUND OUT."



import os
import re
import requests
import time


def replace_redirecting_links(config, logger):
    task = config["current-task"]
    path = task["path-modify-event"]

    regex = re.compile(task["regex"], re.DOTALL)
    logger.debug(f"regex: {regex}")

    updated = False
    match_found = False

    with open(path, "r", encoding="utf-8") as main_file:
        content = main_file.read()

    # logger.debug(f"content: {content}")
    for match in regex.finditer(content):
        match_found = True
        link = match.group(1)
        logger.debug(f"Match found {link}")

        redirect, new_link = check_redirection(link)
        if redirect:
            updated = True
            content = content.replace(link, new_link)
            logger.info(f"Updated link {link} → {new_link}")

    if not match_found:
        logger.debug(f"No redirect links found")
        return False

    if not updated and match_found:
        logger.debug(f"Content is up to date. No links redirected")
        return False

    with open(path, "w", encoding="utf-8") as file:
        time.sleep(3)
        file.write(content)

    return True


def check_redirection(url):
    try:
        response = requests.get(
            url, allow_redirects=False, timeout=10
        )  # Added a timeout of 10 seconds to prevent hanging indefinitely

        # Check if the status code is a redirection (3xx)
        if 300 <= response.status_code < 400:
            return True, response.headers["Location"]
        return False, url

    except requests.RequestException as request_exception:
        raise Exception(f"An error occurred: {request_exception}") from request_exception



from coauthor.utils.c2platform_org import search_using_ddg
from coauthor.utils.c2platform_org import download_html_and_convert_to_markdown
import pytest


def test_search_ddg_nl():
    search_string = "what is c2 platform? site:c2platform.org language:nl"
    results = search_using_ddg(search_string, max_results=1)
    assert results[0]["href"] == "https://c2platform.org/nl/"


def test_search_ddg_en():
    search_string = "what is c2 platform? site:c2platform.org language:en"
    results = search_using_ddg(search_string, max_results=1)
    assert results[0]["href"] == "https://c2platform.org/"


def test_search_ddg_concepts_en():
    search_string = "what are the concepts of c2 platform? site:c2platform.org language:en"
    results = search_using_ddg(search_string, max_results=1)
    assert results[0]["href"] == "https://c2platform.org/docs/concepts/"


def test_search_getting_started():
    search_string = "development environment setup site:c2platform.org"
    results = search_using_ddg(search_string, max_results=1)
    assert results[0]["href"] == "https://c2platform.org/docs/howto/dev-environment/setup/"
    results = search_using_ddg(search_string, max_results=1, return_docs=False)
    assert results[0]["href"] == "https://c2platform.org/docs/howto/dev-environment/setup/"


def test_download_html_and_convert_to_markdown():
    url = "https://c2platform.org/docs/getting-started/"
    getting_started_md = download_html_and_convert_to_markdown(url)
    assert "set up a local development environment" in getting_started_md


def test_download_html_and_convert_to_markdown_exception():
    url = "https://c2platform.org/docs/getting-started-does-not-exist/"

    with pytest.raises(RuntimeError, match="404 Client Error: Not Found for url"):
        download_html_and_convert_to_markdown(url)



def join_with_spaces_and_truncate(text, max_length=100):
    """
    Joins multiline content into a single line using spaces, truncating to the specified max length.

    :param text: The input string with possible newlines.
    :param max_length: The maximum length of the output string (default 100).
    :return: The joined and truncated string.
    """
    lines = text.split("\n")
    joined = " ".join(line.strip() for line in lines if line.strip())
    return joined[:max_length]



import sys

sys.path.insert(0, "/usr/lib/python3/dist-packages")
import dbus
import notify2


def notification(summary, text, icon="notification-message-im"):
    notify2.init("Coauthor")
    notification_instance = notify2.Notification(summary, text, icon)  # Icon name
    notification_instance.show()



import os
from unittest.mock import patch
import tempfile
from coauthor.modules.file_scanner import scan
import logging
from coauthor.utils.logger import Logger
from coauthor.utils.config import get_config


def get_content():
    return f"""
---
categories: [Whatever]
tags: [Whatever, Whatever, Whatever]
title: Whatever
linkTitle: Whatever
translate: true
weight: 2
description: >
  Whatever
---
Whatever
"""


def get_content_nl():
    return f"""
---
categories: [Wat-dan-ook]
tags: [Wat-dan-ook, Wat-dan-ook, Wat-dan-ook]
title: Wat-dan-ook
linkTitle: Wat-dan-ook
weight: 2
description: >
  Wat-dan-ook
---
Wat-dan-ook
"""


def test_process_file_with_openai_agent_translation():
    """
    This test verifies that the translation task can support translations
    using AI, utilizing the 'path_new_replace' attribute to write AI
    content responses to a different file. The functionality is particularly
    suitable for Hugo-based websites, which store translations in distinct
    folders for each language, such as 'en' for English and 'nl' for Dutch.
    """
    path = os.path.join(os.path.dirname(__file__), "data", "coauthor_translation.yml")
    config = get_config(path=path)
    logger = Logger(__name__, level=logging.DEBUG, log_file=f"{os.getcwd()}/debug.log")
    with tempfile.TemporaryDirectory() as temp_dir:
        path_content = os.path.join(temp_dir, "content")
        path = os.path.join(path_content, "en", "whatever", "whatever.md")
        path_nl = os.path.join(path_content, "nl", "whatever", "whatever.md")
        config["workflows"][0]["scan"] = {"filesystem": {"paths": [path_content]}}
        os.makedirs(os.path.dirname(path))
        with open(path, "w") as f:
            f.write(get_content())
        with patch("coauthor.modules.ai.create_chat_completion", return_value=get_content_nl()):
            scan(config, logger)
            task1 = config["workflows"][0]["tasks"][0]
            task2 = config["workflows"][0]["tasks"][1]
            assert task1["id"] == "ai-translate"
            assert task1["response"] == get_content_nl()
            assert task2["id"] == "write-file"

            # Assert that the translated file exists
            assert os.path.exists(path_nl)
            # Assert that the content of the translated file matches expected content
            with open(path_nl, "r") as f:
                content_nl = f.read()
            assert content_nl == get_content_nl()



import os
import logging
import yaml
import re
from coauthor.utils.markdown import get_frontmatter
import time


def process_frontmatter(config, logger):
    """
    Processes a Markdown file's frontmatter and includes content from specified files.

    This function uses get_frontmatter to parse the specified Markdown file. If the frontmatter
    contains an `include-files` property (a list of markdown files), it reads and concatenates
    the content of these files into the task's content.

    :param config: Configuration dictionary containing:
        - 'current-task': contains the current task details.
        - 'current-task'['path-modify-event']: path to the markdown file to process.
    :param logger: Logger instance for logging messages and errors.
    :return: Updated config dictionary with modified content.
    """
    task = config["current-task"]
    path = task["path-modify-event"]

    try:
        frontmatter, frontmatter_str, original_content = get_frontmatter(path)
        include_files = frontmatter.get("include-files", [])

        if include_files:
            combined_content = "---\n"
            combined_content += frontmatter_str + "\n---\n\n"

            current_dir = os.path.dirname(os.path.abspath(path))
            for include_path in include_files:
                full_path = include_path
                if include_path.startswith("./"):
                    full_path = os.path.join(current_dir, include_path[2:])

                try:
                    with open(full_path, "r", encoding="utf-8") as include_file:
                        combined_content += include_file.read() + "\n"
                except FileNotFoundError:
                    logger.warning(f"Include file not found: {full_path}")

            # combined_content = combined_content.rstrip("\n") + "\n"
            task["content"] = combined_content
        else:
            task["content"] = original_content

    except Exception as e:
        logger.error(f"Error processing frontmatter: {str(e)}")
        raise

    time.sleep(3)
    return config



"""Utilities for selecting lines from text content and fetching content from URLs.

This module includes functions to parse and select specific line ranges from
multiline strings and to download text content from URLs, optionally applying
line selection.
"""

import requests


def select_lines(content, range_str):
    """
    Select specific lines from content based on a range string.

    Parameters:
    - content (str): Either a multiline string or a file path to read from.
    - range_str (str): A string specifying lines to select, e.g., "1,2,4-10,15-end".
                       Supports single lines (e.g., "1"), ranges (e.g., "4-10"),
                       and "end" keyword for the last line.

    Returns:
    - str: The selected lines joined by newline characters.
    """
    lines = content.splitlines()
    total_lines = len(lines)
    if total_lines == 0:
        return ""

    selected_indices = set()
    for range_part in range_str.replace(" ", "").split(","):
        if not range_part:
            continue
        try:
            if "-" in range_part:
                start_str, end_str = range_part.split("-")
            else:
                start_str = range_part
                end_str = range_part
            start = int(start_str) - 1
            end = total_lines if end_str.lower() == "end" else int(end_str)
            if start < 0 or end > total_lines:
                raise ValueError(f"Range {range_part} out of bounds (total lines: {total_lines})")
            selected_indices.update(range(start, end))
        except ValueError as value_error:
            raise ValueError(f"Invalid entry {range_part}: {str(value_error)}") from value_error

    selected_lines = [lines[i] for i in sorted(selected_indices)]
    return "\n".join(selected_lines)


def get_url(url, range_str=None):
    """
    Download content from a URL and optionally select specific lines.

    Parameters:
    - url (str): The URL to download the content from.
    - range_str (str, optional): A string specifying lines to select, e.g., "1,2,4-10,15-end".

    Returns:
    - str: The full content or selected lines joined by newline characters.
    """
    response = requests.get(url, timeout=10)  # Set a timeout of 10 seconds
    response.raise_for_status()
    content = response.text
    if range_str:
        return select_lines(content, range_str)
    return content


# pylint: disable=broad-exception-caught
"""
Module for managing project files and executing tools on Git repositories.
"""

import os
import yaml
import subprocess
import shutil
from typing import Dict, List, Any
import importlib.resources
import re
from coauthor.utils.config import get_projects
import urllib.request


def list_tracked_files(project_path: str) -> List[str]:
    """List all tracked files in the Git repository at project_path."""
    try:
        result = subprocess.run(["git", "ls-files"], cwd=project_path, capture_output=True, text=True, check=True)
        if result.returncode != 0:
            raise ValueError(f"Error listing files: {result.stderr}")
        return result.stdout.splitlines()
    except Exception as exception_error:
        return [f"Error: {str(exception_error)}"]


def list_tracked_directories(project_path: str) -> List[str]:
    """List all directories that contain tracked files in the Git repository at project_path."""
    files = list_tracked_files(project_path)
    if files and isinstance(files[0], str) and files[0].startswith("Error:"):
        return files
    directories = set()
    for file in files:
        directory = os.path.dirname(file)
        while directory not in directories:
            directories.add(directory)
            if directory == "":
                break
            directory = os.path.dirname(directory)
    directories = {"." if d == "" else d for d in directories}
    return sorted(directories)


def write_file(project_path: str, path: str, content: str) -> None:
    """Write or update a single file in the project."""
    full_path = os.path.join(project_path, path)
    os.makedirs(os.path.dirname(full_path), exist_ok=True)
    with open(full_path, "w", encoding="utf-8") as f:
        f.write(content)


def write_files(project_path: str, files: List[Dict[str, str]]) -> None:
    """Write or update files in the project."""
    for file in files:
        write_file(project_path, file["path"], file["content"])


def get_files(project_path: str, paths: List[str]) -> Dict[str, str]:
    """Retrieve content of specified files."""
    contents = {}
    for path in paths:
        full_path = os.path.join(project_path, path)
        if os.path.exists(full_path):
            with open(full_path, "r", encoding="utf-8") as f:
                contents[path] = f.read()
        else:
            contents[path] = "File not found"
    return contents


def get_context(project_path: str) -> str:
    result = get_files(project_path, ["COAUTHOR.md"])
    return result.get("COAUTHOR.md", "No context found")


def update_context(project_path: str, content: str) -> Dict[str, str]:
    write_files(project_path, [{"path": "COAUTHOR.md", "content": content}])
    return {"status": "updated"}


def create_directories(project_path: str, directories: List[str]) -> None:
    """Create directories in the project, similar to mkdir -p."""
    for dir_path in directories:
        full_path = os.path.join(project_path, dir_path)
        os.makedirs(full_path, exist_ok=True)


def list_modified_files(project_path: str) -> List[str]:
    """List files with outstanding changes in the specified project path."""
    try:
        result = subprocess.run(
            ["git", "status", "--porcelain"],
            cwd=project_path,
            capture_output=True,
            text=True,
            check=True,
        )
        lines = result.stdout.splitlines()
        modified = []
        for line in lines:
            if line:
                status = line[0:2]
                path = line[3:]
                if status in ("R ", "C "):
                    path = path.split(" -> ")[1]
                modified.append(path)
        return modified
    except Exception as exception_error:
        return [f"Error: {str(exception_error)}"]


def get_diffs(project_path: str, paths: List[str] = []) -> Dict[str, str]:
    """Retrieve diffs for specified files in the project path."""
    try:
        if paths:
            file_list = paths
        else:
            file_list = list_modified_files(project_path)
        if file_list and isinstance(file_list[0], str) and file_list[0].startswith("Error:"):
            return {"error": file_list[0]}
        diffs = {}
        for file in file_list:
            stat_res = subprocess.run(
                ["git", "status", "--porcelain", file],
                cwd=project_path,
                capture_output=True,
                text=True,
                check=False,
            )
            line = stat_res.stdout.strip()
            if not line:
                diffs[file] = "No changes"
                continue
            status = line[0:2]
            if status == "??":
                full_path = os.path.join(project_path, file)
                try:
                    with open(full_path, "r", encoding="utf-8") as f:
                        lines = f.read().splitlines()
                    diff = f"diff --git a/{file} b/{file}\n"
                    diff += "new file mode 100644\n"
                    diff += "index 0000000..e69de29\n"
                    diff += "--- /dev/null\n"
                    diff += f"+++ b/{file}\n"
                    diff += f"@@ -0,0 +1,{len(lines)} @@\n"
                    diff += "\n".join([f"+{l}" for l in lines]) + "\n"
                    diffs[file] = diff
                except Exception as exception_error:
                    diffs[file] = f"Error reading file: {str(exception_error)}"
            else:
                diff_res = subprocess.run(
                    ["git", "diff", "HEAD", file],
                    cwd=project_path,
                    capture_output=True,
                    text=True,
                    check=False,
                )
                diffs[file] = diff_res.stdout or "No changes"
        return diffs
    except Exception as exception_error:
        return {"error": f"Error: {str(exception_error)}"}


def delete_files(project_path: str, paths: List[str]) -> Dict[str, str]:
    """Delete specified files or directories in the project."""
    results = {}
    for path in paths:
        full_path = os.path.join(project_path, path)
        try:
            if os.path.exists(full_path):
                if os.path.isdir(full_path):
                    shutil.rmtree(full_path)
                else:
                    os.remove(full_path)
                results[path] = "Deleted"
            else:
                results[path] = "Not found"
        except Exception as exception_error:
            results[path] = f"Error: {str(exception_error)}"
    return results


def move_files(project_path: str, moves: List[Dict[str, str]]) -> Dict[str, str]:
    """Move files or directories in the project."""
    results = {}
    for move in moves:
        source = os.path.join(project_path, move["source"])
        destination = os.path.join(project_path, move["destination"])
        try:
            if os.path.exists(source):
                os.makedirs(os.path.dirname(destination), exist_ok=True)
                shutil.move(source, destination)
                results[move["source"]] = f'Moved to {move["destination"]}'
            else:
                results[move["source"]] = "Source not found"
        except Exception as exception_error:
            results[move["source"]] = f"Error: {str(exception_error)}"
    return results


def search_files(
    project_path: str, query: str, is_regex: bool = False, context_lines: int = 0
) -> Dict[str, List[Dict[str, Any]]]:
    """Search for a query in tracked files of the project."""
    files = list_tracked_files(project_path)
    if isinstance(files, list) and files and isinstance(files[0], str) and files[0].startswith("Error:"):
        return {"error": [{"message": files[0]}]}
    results: Dict[str, List[Dict[str, Any]]] = {}
    for rel_path in files:
        full_path = os.path.join(project_path, rel_path)
        try:
            with open(full_path, "r", encoding="utf-8") as f:
                lines = f.readlines()
            matches = []
            for i, line in enumerate(lines):
                if (not is_regex and query in line) or (is_regex and re.search(query, line)):
                    match = {
                        "line_number": i + 1,
                        "match_line": line.strip(),
                    }
                    if context_lines > 0:
                        start = max(0, i - context_lines)
                        end = min(len(lines), i + context_lines + 1)
                        match["context"] = [lines[j].strip() for j in range(start, end) if j != i]
                    matches.append(match)
            if matches:
                results[rel_path] = matches
        except Exception as exception_error:
            if "error" not in results:
                results["error"] = []
            results["error"].append({"message": f"{rel_path}: {str(exception_error)}"})
    return results


def run_pytest(project_path: str, test_path: str) -> str:
    """Run Pytest on the specified test file in the project."""
    config_path = os.path.join(project_path, ".coauthor.yml")
    tool_env = ""
    tool_shell = None
    if os.path.exists(config_path):
        with open(config_path, "r", encoding="utf-8") as f:
            config = yaml.safe_load(f)
            tool_env = config.get("tool_environment", "")
            tool_shell = config.get("tool_shell", None)

    cmd = tool_env.strip() + "\n" if tool_env else ""
    cmd += f"pytest {test_path}"

    run_kwargs = {
        "shell": True,
        "cwd": project_path,
        "capture_output": True,
        "text": True,
        "check": True,
    }
    if tool_shell:
        run_kwargs["executable"] = tool_shell

    try:
        result = subprocess.run(cmd, **run_kwargs)
        return result.stdout + result.stderr
    except subprocess.CalledProcessError as exception_error:
        return (
            f"Pytest failed with exit code {exception_error.returncode}:\n"
            f"{exception_error.stdout}{exception_error.stderr}"
        )
    except Exception as exception_error:
        return f"Error executing Pytest: {str(exception_error)}"


def get_url(url: str) -> str:
    """Fetch content from the specified URL."""
    try:
        with urllib.request.urlopen(url) as response:
            content = response.read().decode('utf-8')
            return content
    except Exception as exception_error:
        return f"Error fetching URL: {str(exception_error)}"


def load_tools():
    tools_resource = importlib.resources.files("coauthor.config").joinpath("tools.yml")
    with importlib.resources.as_file(tools_resource) as tools_path:
        with open(tools_path, "r", encoding="utf-8") as f:
            tools_config = yaml.safe_load(f)["tools"]
    return [{"type": "function", "function": t} for t in tools_config]  # Format for OpenAI


def execute_tool(config, tool_name: str, params: Dict, logger) -> Any:
    """Execute a specified tool on a project.

    Args:
        tool_name (str): The name of the tool to execute.
        params (Dict): Parameters for the tool, including 'project_name'.
        logger: Logger instance.

    Returns:
        Any: Result of the tool execution or error dictionary.
    """
    projects = config.get("all_projects") or get_projects(config)
    logger.info(f"Executing tool: {tool_name}, params: {params}")
    # logger.debug(f"Parameters: {params}")
    # logger.debug(f"Projects: {projects}")
    if projects:
        project = next((p for p in projects if p["name"] == params["project_name"]), None)
        if not project:
            raise ValueError(f"Project not found: {params['project_name'] }, projects: {projects}")
        project_path = os.path.expanduser(project.get("path", os.getcwd()))
    else:
        project = config
        project_path = os.path.expanduser(config.get("path", os.getcwd()))

    if project_path == "none":
        raise ValueError(f"File operations are not supported for project {params['project_name']}")

    all_tools = load_tools()
    logger.debug(f"all_tools: {all_tools}")
    project_tools = project.get("tools", all_tools)
    tool_config = next((t for t in project_tools if t["function"]["name"] == tool_name), None)
    if not tool_config:
        raise ValueError(f"Unknown tool: {tool_name}")

    logger.debug(f'Executing tool "{tool_name}" on project "{params["project_name"]}", "{project_path}"')
    if tool_name == "list_tracked_files":
        result = list_tracked_files(project_path)
    elif tool_name == "list_tracked_directories":
        result = list_tracked_directories(project_path)
    elif tool_name == "write_files":
        write_files(project_path, params["files"])
        result = {"status": "success"}
    elif tool_name == "write_file":
        write_file(project_path, params["path"], params["content"])
        result = {"status": "success"}
    elif tool_name == "get_files":
        result = get_files(project_path, params["paths"])
    elif tool_name == "get_context":
        result = get_context(project_path)
    elif tool_name == "update_context":
        result = update_context(project_path, params["content"])
    elif tool_name == "create_directories":
        create_directories(project_path, params["directories"])
        result = {"status": "success"}
    elif tool_name == "list_modified_files":
        result = list_modified_files(project_path)
    elif tool_name == "get_diffs":
        result = get_diffs(project_path, params.get("paths", []))
    elif tool_name == "delete_files":
        result = delete_files(project_path, params["paths"])
    elif tool_name == "move_files":
        result = move_files(project_path, params["moves"])
    elif tool_name == "search_files":
        result = search_files(
            project_path, params["query"], params.get("is_regex", False), params.get("context_lines", 0)
        )
    elif tool_name == "run_pytest":
        result = run_pytest(project_path, params["test_path"])
    elif tool_name == "get_url":
        result = get_url(params["url"])
    else:
        raise ValueError("Unknown tool: {tool_name}")

    if tool_config.get("returns_response", True):
        logger.debug(f"Tool result: {result}")
        return result

    return None



import subprocess
import os


def get_git_diff(file_path):
    """
    Given the path of a file, this function checks if the file is part of a Git repository.
    If it is, the function returns the diff of the file showing outstanding changes.
    If the file is not in a Git repository, it returns None.

    :param file_path: The path to the file.
    :return: The diff of the file or None if not in a Git repository.
    """

    # Check if the file is part of a git repository
    # try:
    # `git rev-parse --is-inside-work-tree` returns `true` if inside repository
    result = subprocess.run(
        ["git", "rev-parse", "--is-inside-work-tree"],
        cwd=os.path.dirname(file_path),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    # if result.returncode != 0 or result.stdout.strip() != b"true":
    #     # Not a git repository
    #     return ""

    # Get the diff of the file
    diff_result = subprocess.run(
        ["git", "diff", file_path], cwd=os.path.dirname(file_path), stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )

    # if diff_result.returncode != 0:
    #     # Could not get a diff, possibly an error occurred
    #     return ""

    return diff_result.stdout.decode("utf-8")

    # except Exception as e:
    #     # Handle unexpected errors, possibly log this
    #     return ""


# Example usage:
# diff = get_git_diff('/path/to/your/file')
# print(diff)



import sys
import os
import tempfile
from unittest.mock import patch
import pytest
from coauthor.main import main
from coauthor import __version__


def test_main(monkeypatch):
    test_args = ["main"]
    with patch.object(sys, "argv", test_args):
        main()
    assert True


def test_version(monkeypatch, capsys):
    test_args = ["main", "--version"]
    with pytest.raises(SystemExit) as exc_info:
        with patch.object(sys, "argv", test_args):
            main()
    assert exc_info.value.code == 0
    captured = capsys.readouterr()
    assert captured.out.startswith(f"coauthor {__version__}")


def test_main_with_config_path_that_does_not_exist(monkeypatch):
    test_args = ["main", "--config_path", "test_config.yaml", "--profile", "test_profile"]
    with patch.object(sys, "argv", test_args):
        with pytest.raises(FileNotFoundError):
            main()


def test_main_with_config_path(monkeypatch):
    file_path_expected = os.path.join(os.getcwd(), ".coauthor.yml")
    test_args = ["main", "--config_path", file_path_expected]
    with patch.object(sys, "argv", test_args):
        monkeypatch.setattr("builtins.print", lambda x: x)
        main()


def test_main_with_no_arguments(monkeypatch):
    test_args = ["main"]
    with patch.object(sys, "argv", test_args):
        monkeypatch.setattr("builtins.print", lambda x: x)
        main()


def test_main_log_file(monkeypatch):
    with tempfile.TemporaryDirectory() as temp_dir:
        test_args = ["main", "--log_file", f"{os.getcwd()}/debug.log"]
        with patch.object(sys, "argv", test_args):
            main()


# def test_script_execution():
#     script_path = os.path.join(os.getcwd(), "src", "coauthor", "main.py")
#     python_interpreter = os.path.expanduser("~/.virtualenv/c2d_ai/bin/python")
#     src_dir = os.path.join(os.getcwd(), "src")
#     env = os.environ.copy()
#     env["PYTHONPATH"] = src_dir
#     result = subprocess.run([python_interpreter, script_path], capture_output=True, text=True, env=env)
#     assert result.returncode == 0



import requests
import pytest
from unittest import mock, TestCase
import re
from coauthor.modules.web_tasks import check_redirection, replace_redirecting_links


def test_check_redirection_redirect():
    url = "https://oreil.ly/gKq4m"
    expected_location = "https://kubernetes.io/docs/setup/production-environment/tools/kubeadm/install-kubeadm/"
    with mock.patch("requests.get") as mock_get:
        mock_response = mock.Mock()
        mock_response.status_code = 301
        mock_response.headers = {"Location": expected_location}
        mock_get.return_value = mock_response

        success, location = check_redirection(url)
        assert success is True
        assert location == expected_location


def test_check_redirection_no_redirect():
    url = "https://c2platform.org/"
    with mock.patch("requests.get") as mock_get:
        mock_response = mock.Mock()
        mock_response.status_code = 200
        mock_response.headers = {}
        mock_get.return_value = mock_response

        success, location = check_redirection(url)
        assert success is False
        assert location == url


def test_check_redirection_exception():
    url = "https://invalid-url"
    with mock.patch("requests.get", side_effect=requests.RequestException("Error occurred")) as mock_get:
        with pytest.raises(Exception) as excinfo:
            check_redirection(url)
        assert "An error occurred: Error occurred" in str(excinfo.value)


def test_check_redirection_real_redirect():
    url = "https://httpbin.org/absolute-redirect/1"
    expected_location = "http://httpbin.org/get"

    success, location = check_redirection(url)
    assert success is True
    assert location == expected_location

    url = "https://httpbin.org/absolute-redirect/2"
    expected_location = "http://httpbin.org/absolute-redirect/1"

    success, location = check_redirection(url)
    assert success is True
    assert location == expected_location


def test_check_redirection_real_no_redirect():
    url = "https://httpbin.org/get"

    success, location = check_redirection(url)
    assert success is False
    assert location == url


def test_check_redirection_real_redirect_oreilly():
    url = "https://oreil.ly/gKq4m"
    expected_location = "https://kubernetes.io/docs/setup/production-environment/tools/kubeadm/install-kubeadm/"

    success, location = check_redirection(url)
    assert success is True
    assert location == expected_location


def test_replace_redirecting_links_no_match():
    config_mock = {"current-task": {"path-modify-event": "test_path", "regex": r"(https?://[^\s]+)"}}
    logger_mock = mock.Mock()
    mock_open = mock.mock_open(read_data="No links here")

    with mock.patch("builtins.open", mock_open):
        result = replace_redirecting_links(config_mock, logger_mock)

    assert result is False
    logger_mock.debug.assert_any_call("No redirect links found")


def test_replace_redirecting_links_found_and_replaced():
    config_mock = {"current-task": {"path-modify-event": "test_path", "regex": r"(https?://[^\s]+)"}}
    logger_mock = mock.Mock()
    mock_open = mock.mock_open(read_data="Check this link: http://example.com/old")

    with mock.patch("builtins.open", mock_open):
        with mock.patch("coauthor.modules.web_tasks.check_redirection", return_value=(True, "http://example.com/new")):
            with mock.patch("time.sleep", return_value=None):
                result = replace_redirecting_links(config_mock, logger_mock)

    assert result is True
    mock_open().write.assert_called_with("Check this link: http://example.com/new")
    logger_mock.info.assert_any_call("Updated link http://example.com/old → http://example.com/new")


def test_replace_redirecting_links_found_not_replaced():
    config_mock = {"current-task": {"path-modify-event": "test_path", "regex": r"(https?://[^\s]+)"}}
    logger_mock = mock.Mock()
    mock_open = mock.mock_open(read_data="Check this link: http://example.com/same")

    with mock.patch("builtins.open", mock_open):
        with mock.patch(
            "coauthor.modules.web_tasks.check_redirection", return_value=(False, "http://example.com/same")
        ):
            result = replace_redirecting_links(config_mock, logger_mock)

    assert result is False
    logger_mock.debug.assert_any_call("Content is up to date. No links redirected")



from coauthor.utils.logger import Logger
import logging
import os


def test_logger_default_log_file():
    """
    Test that the Logger class creates a default log file named 'coauthor.log' in the current working directory.

    The test initializes the Logger with INFO level logging. It then cleans up any preexisting log file named
    'coauthor.log'. The test asserts that the logger's log file is correctly set to 'coauthor.log' located in the current
    working directory.
    """
    logger = Logger(__name__, level=logging.INFO)
    log_file = os.path.join(os.getcwd(), "coauthor.log")
    logger.clean_log_file()
    assert logger.log_file == log_file


def test_logger_default_log_file_removed():
    """
    Test the Logger class behavior when a default log file is missing.

    This test initializes the Logger with INFO level logging, removes any existing 'coauthor.log' file from the
    current working directory, and invokes the clean_log_file method on the logger. It asserts that the logger
    correctly identifies and sets the log file to 'coauthor.log' after cleaning.
    """
    logger = Logger(__name__, level=logging.INFO)
    log_file = os.path.join(os.getcwd(), "coauthor.log")
    os.remove(log_file)
    logger.clean_log_file()
    assert logger.log_file == log_file


def test_logger_log_file():
    """
    Test the Logger with a custom log file path.

    The test checks whether the logger correctly logs messages to a specified file path. It initializes the
    Logger with INFO level logging and a custom log file located at '/tmp/coauthor.log'. After logging a test
    message, it verifies the existence of the log file and asserts that the Logger's log file path is set as
    specified and that the file exists.
    """
    log_file = "/tmp/coauthor.log"
    logger = Logger(__name__, level=logging.INFO, log_file=log_file)
    logger.info("hi ho")
    file_exists = os.path.exists(log_file)
    os.remove(log_file)
    assert [logger.log_file, file_exists] == [log_file, True]


def test_logger_log_file_debug():
    """
    Test the Logger's capability to handle DEBUG level logs to a specified file path.

    This test initializes the Logger with DEBUG level logging and a custom log file at '/tmp/coauthor.log'. It sends
    debug and error level messages to the logger, checks for the existence of the log file, and then asserts
    that the logger's file path configuration and file creation are correct without removing it.
    """
    log_file = "/tmp/coauthor.log"
    logger = Logger(__name__, level=logging.DEBUG, log_file=log_file)
    logger.debug("hi ho")
    logger.error("hi ho")
    file_exists = os.path.exists(log_file)
    # os.remove(log_file)
    assert [logger.log_file, file_exists] == [log_file, True]



import pytest
from unittest import mock
from unittest.mock import mock_open
import os

# import subprocess
# import requests
from coauthor.utils.logger import Logger
import logging
from coauthor.modules.plantuml import download_plantuml_jar, process_puml_file


@pytest.fixture
def mock_config():
    return {
        "plantuml": {"url": "https://example.com/plantuml.jar", "path": "/tmp/plantuml.jar"},
        "current-task": {"path-modify-event": "/tmp/sample.puml"},
    }


@pytest.fixture
def mock_logger():
    return mock.Mock()
    # return Logger(__name__, level=logging.DEBUG, log_file=f"{os.getcwd()}/debug.log")


def test_download_plantuml_jar(mock_config, mock_logger):
    with mock.patch("requests.get") as mock_get, mock.patch("builtins.open", mock_open()) as mock_file, mock.patch(
        "os.path.exists", return_value=False
    ):

        mock_response = mock.Mock()
        mock_response.content = b"jar content"
        mock_get.return_value = mock_response

        download_plantuml_jar(mock_config, mock_logger)

        mock_get.assert_called_once_with("https://example.com/plantuml.jar", timeout=30)
        mock_file().write.assert_called_once_with(b"jar content")
        mock_logger.info.assert_called_once_with("Downloaded PlantUML jar to /tmp/plantuml.jar")


def test_download_plantuml_jar_file_exists(mock_config, mock_logger):
    with mock.patch("os.path.exists", return_value=True):
        download_plantuml_jar(mock_config, mock_logger)

        mock_logger.info.assert_not_called()


def test_process_puml_file(mock_config, mock_logger):
    with mock.patch("subprocess.run") as mock_subproc_run, mock.patch("os.path.exists", return_value=True):

        process_puml_file(mock_config, mock_logger)

        mock_subproc_run.assert_any_call(["java", "-jar", "/tmp/plantuml.jar", "-tsvg", "/tmp/sample.puml"], check=True)
        mock_subproc_run.assert_any_call(["java", "-jar", "/tmp/plantuml.jar", "-tpng", "/tmp/sample.puml"], check=True)
        mock_logger.info.assert_any_call("Exported /tmp/sample.puml to svg, png")
